from pomagma.util import DB
from pomagma.util import suggest_region_sizes
import glob
import itertools
import multiprocessing
import os
from parsable import parsable
import pomagma.atlas
import pomagma.cartographer
import pomagma.surveyor
import pomagma.theorist
import pomagma.util
import shutil
import subprocess
import sys
import time

parsable = parsable.Parsable()

DEFAULT_SURVEY_SIZE = 16384 + 512 - 1
MIN_SLEEP_SEC = 1
MAX_SLEEP_SEC = 600
PYTHON = sys.executable


class parsable_fork(object):

    def __init__(self, fun, *args, **kwargs):
        self.args = [PYTHON, '-m', 'pomagma.workers', fun.__name__]
        self.args += map(str, args)
        for key, val in kwargs.iteritems():
            self.args.append('{}={}'.format(key, val))
        self.proc = subprocess.Popen(self.args)

    def wait(self):
        self.proc.wait()
        code = self.proc.returncode
        assert code == 0, '\n'.join([
            'forked command failed with exit code {}'.format(code),
            ' '.join(self.args)])

    def terminate(self):
        if self.proc.poll() is None:
            self.proc.terminate()


class fork(object):

    def __init__(self, fun, *args, **kwargs):
        self.command = '{}({})'.format(fun.__name__, ', '.join([
            str(arg) for arg in args
        ] + [
            '{}={}'.format(key, repr(val)) for key, val in kwargs.iteritems()
        ]))
        self.proc = multiprocessing.Process(
            target=fun,
            args=args,
            kwargs=kwargs)
        self.proc.start()

    def wait(self):
        self.proc.join()
        code = self.proc.exitcode
        assert code == 0, '\n'.join([
            'forked command failed with exit code {}'.format(code),
            self.command])

    def terminate(self):
        if self.proc.is_alive():
            self.proc.terminate()


class Sleeper(object):

    def __init__(self, name):
        self.name = name
        self.duration = MIN_SLEEP_SEC

    def reset(self):
        self.duration = MIN_SLEEP_SEC

    def sleep(self):
        sys.stderr.write('# {} sleeping\n'.format(self.name))
        sys.stderr.flush()
        time.sleep(self.duration)
        self.duration = min(MAX_SLEEP_SEC, 2 * self.duration)


class FileQueue(object):

    def __init__(self, path, template='{}'):
        self.path = path
        self.template = template
        self.pattern = os.path.join(self.path, DB(template.format('[0-9]*')))

    def get(self):
        # specifically ignore temporary files like temp.1234.0.pb
        return glob.glob(self.pattern)

    def __iter__(self):
        return iter(self.get())

    def __len__(self):
        return len(self.get())

    def try_pop(self, destin):
        for source in self:
            os.rename(source, destin)
            return True
        return False

    def push(self, source):
        if self.path and not os.path.exists(self.path):
            os.makedirs(self.path)
        with pomagma.util.mutex(self.path):
            for i in itertools.count():
                destin = os.path.join(self.path, DB(self.template.format(i)))
                if not os.path.exists(destin):
                    os.rename(source, destin)
                    return

    def clear(self):
        for item in self:
            os.remove(item)


class CartographerWorker(object):

    def __init__(self, theory, region_size, region_queue_size, **options):
        self.options = options
        self.log_file = options['log_file']
        self.world = DB('world')
        self.normal_world = DB('world.normal')
        self.normal_region = DB('region.normal.{:d}')
        self.min_size = pomagma.util.MIN_SIZES[theory]
        self.region_size = region_size
        self.region_queue = FileQueue('region.queue')
        self.survey_queue = FileQueue('survey.queue')
        self.region_queue_size = region_queue_size
        self.diverge_conjectures = 'diverge_conjectures.facts'
        self.diverge_theorems = 'diverge_theorems.facts'
        self.equal_conjectures = 'equal_conjectures.facts'
        DEBUG = False
        if DEBUG:
            options = pomagma.util.use_memcheck(
                options,
                'cartographer.memcheck.out')
        self.server = pomagma.cartographer.serve(theory, self.world, **options)
        self.db = self.server.connect()
        self.infer_state = 0
        if os.path.exists(self.normal_world):
            world_digest = pomagma.atlas.get_hash(self.world)
            normal_world_digest = pomagma.atlas.get_hash(self.normal_world)
            if world_digest == normal_world_digest:
                self.infer_state = 2

    def stop(self):
        self.db.stop()
        # self.server.stop()

    def log(self, message):
        rss = pomagma.util.get_rss(self.server.pid)
        message = 'Cartographer {}k {}'.format(rss, message)
        pomagma.util.log_print(message, self.log_file)

    def is_normal(self):
        assert self.infer_state in [0, 1, 2]
        return self.infer_state == 2

    def garbage_collect(self):
        # assume surveyor.dump takes < 1.0 days
        pomagma.atlas.garbage_collect(grace_period_days=1.0)

    def try_work(self):
        return (
            self.try_produce_regions() or
            self.try_normalize() or
            self.try_consume_surveys()
        )

    def try_produce_regions(self):
        queue_size = len(self.region_queue)
        if queue_size >= self.region_queue_size:
            return False
        else:
            self.fill_region_queue(self.region_queue)
            return True

    def try_normalize(self):
        if self.is_normal():
            return False
        else:
            self.log('Inferring {}'.format(['pos', 'neg'][self.infer_state]))
            if self.db.infer(self.infer_state):
                self.db.validate()
                self.db.dump(self.world)
                self.garbage_collect()
                self.replace_region_queue()
            else:
                self.infer_state += 1
                if self.is_normal():
                    self.log('Normalized')
                    self.db.dump(self.normal_world)
                    self.trim_normal_regions()
                    self.garbage_collect()
                    self.theorize()
            return True

    def try_consume_surveys(self):
        surveys = self.survey_queue.get()
        if not surveys:
            return False
        else:
            self.log('Aggregating {} surveys'.format(len(surveys)))
            for survey in surveys:
                self.db.aggregate(survey)
                self.db.validate()
                self.db.dump(self.world)
                self.garbage_collect()
                self.infer_state = 0
                world_size = self.db.info()['item_count']
                self.log('world_size = {}'.format(world_size))
                os.remove(survey)
            self.db.crop()
            self.replace_region_queue()
            return True

    def fill_region_queue(self, queue):
        self.log('Filling region queue')
        if not os.path.exists(queue.path):
            os.makedirs(queue.path)
        queue_size = len(queue)
        trim_count = max(0, self.region_queue_size - queue_size)
        regions_out = []
        for i in itertools.count():
            region_out = os.path.join(queue.path, DB(i))
            if not os.path.exists(region_out):
                regions_out.append(region_out)
                if len(regions_out) == trim_count:
                    break
        # trim in parallel because these are small
        self.db.trim([
            {'size': self.region_size, 'filename': r}
            for r in regions_out
        ])

    def replace_region_queue(self):
        self.log('Replacing region queue')
        with pomagma.util.temp_copy(self.region_queue.path) as temp_path:
            self.fill_region_queue(FileQueue(temp_path))
            self.region_queue.clear()
            self.garbage_collect()

    def trim_normal_regions(self):
        self.log('Trimming normal regions')
        assert self.is_normal()
        max_size = self.db.info()['item_count']
        # trim sequentially because these are large
        for size in suggest_region_sizes(self.min_size, max_size):
            filename = self.normal_region.format(size)
            self.db.trim([{'size': size, 'filename': filename}])

    def theorize(self):
        self.log('Theorizing')
        conjectures = self.diverge_conjectures
        theorems = self.diverge_theorems
        self.db.conjecture(conjectures, self.equal_conjectures)
        with pomagma.util.temp_copy(conjectures) as temp_conjectures:
            with pomagma.util.temp_copy(theorems) as temp_theorems:
                if os.path.exists(theorems):
                    shutil.copyfile(theorems, temp_theorems)
                theorem_count = pomagma.theorist.try_prove_diverge(
                    conjectures,
                    temp_conjectures,
                    temp_theorems,
                    **self.options)
        if theorem_count > 0:
            self.log('Proved {} theorems'.format(theorem_count))
            counts = self.db.assume(theorems)
            if counts['pos'] + counts['neg']:
                self.log('Assumed {} pos + {} neg facts'.format(
                    counts['pos'],
                    counts['neg']))
                self.db.validate()
                self.db.dump(self.world)
                self.garbage_collect()
                self.infer_state = 0 if counts['pos'] else 1
                self.replace_region_queue()


@parsable
def cartographer_work(
        theory,
        region_size=(DEFAULT_SURVEY_SIZE - 512),
        region_queue_size=4,
        **options):
    '''
    Start cartographer worker.
    '''
    min_size = pomagma.util.MIN_SIZES[theory]
    assert region_size >= min_size
    options.setdefault('log_file', 'cartographer.log')
    with pomagma.atlas.chdir(theory), pomagma.util.mutex(DB('world')):
        worker = CartographerWorker(
            theory,
            region_size,
            region_queue_size,
            **options)
        try:
            sleeper = Sleeper('cartographer')
            while True:
                if not worker.try_work():
                    sleeper.sleep()
                else:
                    sleeper.reset()
        finally:
            worker.stop()


def cartographer(*args, **kwargs):
    return parsable_fork(cartographer_work, *args, **kwargs)


@parsable
def surveyor_work(theory, step_size=512, **options):
    '''
    Start surveyor worker.
    '''
    assert step_size > 0
    with pomagma.atlas.chdir(theory):
        region_queue = FileQueue('region.queue')
        survey_queue = FileQueue('survey.queue')
        region = pomagma.util.temp_name(DB('region'))
        survey = pomagma.util.temp_name(DB('survey'))
        options.setdefault('log_file', 'survey.log')
        sleeper = Sleeper('surveyor')
        while True:
            if not region_queue.try_pop(region):
                sleeper.sleep()
            else:
                sleeper.reset()
                region_size = pomagma.atlas.get_item_count(region)
                survey_size = region_size + step_size
                pomagma.surveyor.survey(
                    theory,
                    region,
                    survey,
                    survey_size,
                    **options)
                os.remove(region)
                survey_queue.push(survey)


def surveyor(*args, **kwargs):
    return parsable_fork(surveyor_work, *args, **kwargs)


if __name__ == '__main__':
    parsable()



from pomagma.compiler import compiler
from pomagma.compiler import signature
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import NotNegatable
from pomagma.compiler.expressions import try_negate_name
from pomagma.compiler.plans import add_costs
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.util import methodof

MIN_SPLIT_COST = 1.5  # above which we split the outermost for loop


def get_set_var(name, args):
    return '_'.join([name] + [a.var.name.rstrip('_') for a in args])


@methodof(compiler.Iter, 'program')
def Iter_program(self, program, stack=None, poll=None):
    pos_sets = []
    neg_sets = []
    for_ = 'FOR_ALL {val}'.format(val=self.var)

    for test in self.tests:
        if test.name == 'UNKNOWN':
            modifier = 'UNKNOWN'
            test = test.args[0]
        else:
            modifier = 'POS'
        set_var = get_set_var(test.name, test.args)
        if modifier == 'POS' and test.arity == 'UnaryRelation':
            for_ = 'FOR_UNARY_RELATION {rel} {val}'.format(
                rel=test.name,
                val=self.var)
            pos_set = 'LETS_UNARY_RELATION {rel} {var}'.format(
                rel=test.name,
                var=set_var)
            pos_sets.append((set_var, pos_set))
        elif modifier == 'POS' and test.arity == 'BinaryRelation':
            lhs, rhs = test.args
            assert lhs != rhs, lhs
            PARITY = ('LHS' if self.var == rhs else 'RHS')
            for_ = 'FOR_BINARY_RELATION_{PARITY} {rel} {lhs} {rhs}'.format(
                PARITY=PARITY,
                rel=test.name,
                lhs=lhs,
                rhs=rhs)
            pos_set = 'LETS_BINARY_RELATION_{PARITY} {rel} {lhs} {rhs}'.format(
                PARITY=PARITY,
                rel=test.name,
                lhs=set_var if self.var == lhs else lhs,
                rhs=set_var if self.var == rhs else rhs)
            pos_sets.append((set_var, pos_set))
        elif modifier == 'UNKNOWN' and test.arity == 'UnaryRelation':
            for_ = None
            neg_set = 'LETS_UNARY_RELATION {rel} {var}'.format(
                rel=test.name,
                var=set_var)
            neg_sets.append((set_var, neg_set))
        elif modifier == 'UNKNOWN' and test.arity == 'BinaryRelation':
            for_ = None
            lhs, rhs = test.args
            assert lhs != rhs, lhs
            PARITY = ('LHS' if self.var == rhs else 'RHS')
            neg_set = \
                'LETS_BINARY_RELATION_{PARITY} {rel} {lhs} {rhs}'.format(
                    PARITY=PARITY,
                    rel=test.name,
                    lhs=set_var if self.var == lhs else lhs,
                    rhs=set_var if self.var == rhs else rhs)
            neg_sets.append((set_var, neg_set))
        else:
            raise ValueError(
                'invalid modifier,arity: {},{}'.format(modifier, test.arity))

    for var, expr in sorted(self.lets.iteritems()):
        assert self.var in expr.args,\
            '{} not in {}'.format(self.var, expr.args)
        set_var = expr.var.name
        if expr.arity == 'InjectiveFunction':
            for_ = 'FOR_INJECTIVE_FUNCTION {fun} {key} {val}'.format(
                fun=expr.name,
                key=self.var,
                val=var)
            pos_set = 'LETS_INJECTIVE_FUNCTION {fun} {var}'.format(
                fun=expr.name,
                var=set_var)
        elif expr.arity == 'BinaryFunction':
            lhs, rhs = expr.args
            assert lhs != rhs, lhs
            if self.var == lhs:
                for_ = \
                    'FOR_BINARY_FUNCTION_RHS {fun} {lhs} {rhs} {val}'.format(
                        fun=expr.name,
                        lhs=lhs,
                        rhs=rhs,
                        val=expr.var.name)
                pos_set = 'LETS_BINARY_FUNCTION_RHS {fun} {lhs} {rhs}'.format(
                    fun=expr.name,
                    lhs=set_var,
                    rhs=rhs)
            else:
                for_ = \
                    'FOR_BINARY_FUNCTION_LHS {fun} {lhs} {rhs} {val}'.format(
                        fun=expr.name,
                        lhs=lhs,
                        rhs=rhs,
                        val=expr.var.name)
                pos_set = 'LETS_BINARY_FUNCTION_LHS {fun} {lhs} {rhs}'.format(
                    fun=expr.name,
                    lhs=lhs,
                    rhs=set_var)
        elif expr.arity == 'SymmetricFunction':
            lhs, rhs = expr.args
            assert lhs != rhs, lhs
            fixed = rhs if self.var == lhs else lhs
            for_ = 'FOR_SYMMETRIC_FUNCTION_LHS {fun} {lhs} {rhs} {val}'.format(
                fun=expr.name,
                lhs=fixed,
                rhs=self.var,
                val=expr.var.name)
            pos_set = 'LETS_SYMMETRIC_FUNCTION_LHS {fun} {lhs} {rhs}'.format(
                fun=expr.name,
                lhs=fixed,
                rhs=set_var)
        else:
            raise ValueError('invalid arity {}'.format(expr.arity))
        pos_sets.append((set_var, pos_set))

    if len(pos_sets) <= 1 and len(neg_sets) == 0:
        program.append(for_)
        if poll:
            program.append('IF_BLOCK {val}'.format(val=self.var))
    else:
        for_ = 'FOR{pos}{neg} {val}'.format(
            pos='_POS' * len(pos_sets),
            neg='_NEG' * len(neg_sets),
            val=self.var)
        for set_var, pos_set in pos_sets + neg_sets:
            program.append(pos_set)
            for_ += ' {var}'.format(var=set_var)
        program.append(for_)
        if poll:
            program.append('IF_BLOCK {val}'.format(val=self.var))
        for var, expr in sorted(self.lets.iteritems()):
            arity = expr.arity
            ARITY = arity.replace('Function', '').upper()
            assert ARITY in ['INJECTIVE', 'BINARY', 'SYMMETRIC'], ARITY
            args = [a.name for a in expr.args]
            line = 'LET_{ARITY}_FUNCTION {fun} {args} {val}'.format(
                ARITY=ARITY,
                fun=expr.name,
                args=' '.join(args),
                val=var)
            program.append(line)
    self.body.program(program, stack=self.stack)


@methodof(compiler.IterInvInjective, 'program')
def IterInvInjective_program(self, program, stack=None, poll=None):
    program.append('FOR_INJECTIVE_FUNCTION_INVERSE {fun} {key} {val}'.format(
        fun=self.fun,
        key=self.var,
        val=self.value))
    self.body.program(program, poll=poll)


@methodof(compiler.IterInvBinary, 'program')
def IterInvBinary_program(self, program, stack=None, poll=None):
    ARITY = signature.get_arity(self.fun).replace('Function', '').upper()
    program.append('FOR_{ARITY}_FUNCTION_VAL {fun} {lhs} {rhs} {val}'.format(
        ARITY=ARITY,
        fun=self.fun,
        lhs=self.var1,
        rhs=self.var2,
        val=self.value))
    if poll:
        raise NotImplementedError('cannot poll IterInvBinary')
        # program.append('IF_BLOCK {val}'.format(val=self.var1))  # arbitrary
    self.body.program(program)


@methodof(compiler.IterInvBinaryRange, 'program')
def IterInvBinaryRange_program(self, program, stack=None, poll=None):
    PARITY = ('LHS' if self.lhs_fixed else 'RHS')
    ARITY = signature.get_arity(self.fun).replace('Function', '').upper()

    line = 'FOR_{ARITY}_FUNCTION_{PARITY}_VAL {fun} {lhs} {rhs} {val}'.format(
        ARITY=ARITY,
        PARITY=PARITY,
        fun=self.fun,
        lhs=self.var1,
        rhs=self.var2,
        val=self.value)
    program.append(line)
    if poll:
        raise NotImplementedError('cannot poll IterInvBinaryRange')
        # program.append('IF_BLOCK {val}'.format(val=var))  # arbitrary
    self.body.program(program)


@methodof(compiler.Let, 'program')
def Let_program(self, program, stack=None, poll=None):
    if not (stack and self in stack):
        arity = self.expr.arity
        args = [arg.name for arg in self.expr.args]
        if arity == 'NullaryFunction':
            line = 'FOR_NULLARY_FUNCTION {fun} {val}'.format(
                fun=self.expr.name,
                val=self.var)
        elif arity == 'InjectiveFunction':
            line = 'FOR_INJECTIVE_FUNCTION_KEY {fun} {key} {val}'.format(
                fun=self.expr.name,
                key=args[0],
                val=self.var)
        elif arity == 'BinaryFunction':
            line = 'FOR_BINARY_FUNCTION_LHS_RHS {fun} {lhs} {rhs} {val}'\
                .format(
                    fun=self.expr.name,
                    lhs=args[0],
                    rhs=args[1],
                    val=self.var)
        elif arity == 'SymmetricFunction':
            line = 'FOR_SYMMETRIC_FUNCTION_LHS_RHS {fun} {lhs} {rhs} {val}'\
                .format(
                    fun=self.expr.name,
                    lhs=args[0],
                    rhs=args[1],
                    val=self.var)
        else:
            raise ValueError('unknown arity: {}'.format(arity))
        program.append(line)
    self.body.program(program, stack=stack, poll=poll)


@methodof(compiler.Test, 'program')
def Test_program(self, program, stack=None, poll=None):
    if not (stack and self in stack):
        arity = self.expr.arity
        args = [arg.name for arg in self.expr.args]
        if self.expr.name == 'EQUAL':
            line = 'IF_EQUAL {lhs} {rhs}'.format(
                lhs=self.expr.args[0],
                rhs=self.expr.args[1])
        elif arity == 'UnaryRelation':
            line = 'IF_UNARY_RELATION {rel} {key}'.format(
                rel=self.expr.name,
                key=args[0])
        elif arity == 'BinaryRelation':
            line = 'IF_BINARY_RELATION {rel} {lhs} {rhs}'.format(
                rel=self.expr.name,
                lhs=args[0],
                rhs=args[1])
        elif arity == 'NullaryFunction':
            line = 'IF_NULLARY_FUNCTION {fun} {val}'.format(
                fun=self.expr.name,
                val=self.expr.var.name)
        elif arity == 'InjectiveFunction':
            line = 'IF_INJECTIVE_FUNCTION {fun} {key} {val}'.format(
                fun=self.expr.name,
                key=args[0],
                val=self.expr.var.name)
        elif arity == 'BinaryFunction':
            line = 'IF_BINARY_FUNCTION {fun} {lhs} {rhs} {val}'.format(
                fun=self.expr.name,
                lhs=args[0],
                rhs=args[1],
                val=self.expr.var.name)
        elif arity == 'SymmetricFunction':
            line = 'IF_SYMMETRIC_FUNCTION {fun} {lhs} {rhs} {val}'.format(
                fun=self.expr.name,
                lhs=args[0],
                rhs=args[1],
                val=self.expr.var.name)
        else:
            raise ValueError('unknown arity: {}'.format(arity))
        program.append(line)
    self.body.program(program, stack=stack, poll=poll)


@methodof(compiler.Ensure, 'program')
def Ensure_program(self, program, stack=None, poll=None):
    expr = self.expr
    args = [arg if arg.args else arg.var for arg in expr.args]
    if all(arg.is_var() for arg in args):

        arity = self.expr.arity
        if self.expr.name == 'EQUAL':
            line = 'INFER_EQUAL {lhs} {rhs}'.format(
                lhs=args[0],
                rhs=args[1])
        elif arity == 'UnaryRelation':
            line = 'INFER_UNARY_RELATION {rel} {key}'.format(
                rel=self.expr.name,
                key=args[0])
        elif arity == 'BinaryRelation':
            line = 'INFER_BINARY_RELATION {rel} {lhs} {rhs}'.format(
                rel=self.expr.name,
                lhs=args[0],
                rhs=args[1])
        else:
            raise ValueError('unknown arity: {}'.format(arity))

    elif any(arg.is_var() for arg in args):
        assert self.expr.name == 'EQUAL', self.expr.name

        if args[0].is_var():
            var, expr = args
        else:
            expr, var = args
        args = [arg.var.name for arg in expr.args]
        ARITY = expr.arity.replace('Function', '').upper()
        line = 'INFER_{ARITY}_FUNCTION {fun} {args} {var}'.format(
            ARITY=ARITY,
            fun=expr.name,
            args=' '.join(args),
            var=var.name)
    else:
        assert self.expr.name == 'EQUAL', self.expr.name

        lhs, rhs = args
        if (len(rhs.args), rhs.name) < (len(lhs.args), lhs.name):
            lhs, rhs = rhs, lhs
        line = 'INFER_{ARITY1}_{ARITY2} {fun1} {args1} {fun2} {args2}'.format(
            ARITY1=lhs.arity.replace('Function', '').upper(),
            ARITY2=rhs.arity.replace('Function', '').upper(),
            fun1=lhs.name,
            fun2=rhs.name,
            args1=' '.join(a.var.name for a in lhs.args),
            args2=' '.join(a.var.name for a in rhs.args))

    program.append(line)


def write_full_programs(programs, sequents, can_parallelize=True):
    full_tasks = []
    for sequent in sequents:
        for cost, seq, plan in compiler.compile_full(sequent):
            full_tasks.append((cost, sequent, seq, plan))
    full_tasks.sort()
    for plan_id, (cost, sequent, seq, plan) in enumerate(full_tasks):
        poll = can_parallelize and (cost >= MIN_SPLIT_COST)
        programs += [
            '',
            '# plan {}: cost = {:0.1f}'.format(plan_id, cost),
            '# using {}'.format(sequent),
            '# infer {}'.format(seq),
        ]
        if poll:
            programs.append('FOR_BLOCK')
        plan.program(programs, poll=poll)


def write_event_programs(programs, sequents):

    event_tasks = {}
    for sequent in sequents:
        for event in compiler.get_events(sequent):
            name = 'Variable' if event.is_var() else event.name
            plans = sorted(compiler.compile_given(sequent, event))
            cost = add_costs(c for (c, _, _) in plans)
            tasks = event_tasks.setdefault(name, [])
            tasks.append((cost, event, sequent, plans))

    group_tasks = {}
    for name, tasks in event_tasks.iteritems():
        groupname = signature.get_arity(name)
        group_tasks.setdefault(groupname, {})[name] = sorted(tasks)

    group_tasks = sorted(group_tasks.iteritems())
    group_id = 0
    for groupname, group in group_tasks:
        group = sorted(group.iteritems())
        arity = signature.get_arity(group[0][0])

        for eventname, tasks in group:
            total_cost = add_costs(c for (c, _, _, _) in tasks)
            programs += [
                '',
                '# ' + '-' * 76,
                '# plans {}.*: total cost = {:0.1f}'.format(
                    group_id,
                    total_cost),
                '# given {}'.format(eventname),
            ]

            plan_id = 0
            for _, event, sequent, plans in tasks:
                diagonal = (
                    len(event.args) == 2 and event.args[0] == event.args[1])
                if diagonal:
                    lhs = event.args[0]
                    assert lhs.arity == 'Variable'
                    rhs = Expression.make(lhs.name + '_')
                    event = Expression.make(event.name, lhs, rhs)

                if arity == 'Variable':
                    given = 'GIVEN_EXISTS {var}'.format(var=event.name)
                elif arity == 'UnaryRelation':
                    given = 'GIVEN_UNARY_RELATION {rel} {key}'.format(
                        rel=event.name,
                        key=event.args[0])
                elif arity == 'BinaryRelation':
                    given = 'GIVEN_BINARY_RELATION {rel} {lhs} {rhs}'.format(
                        rel=event.name,
                        lhs=event.args[0],
                        rhs=event.args[1])
                elif arity == 'NullaryFunction':
                    given = 'GIVEN_NULLARY_FUNCTION {fun} {val}'\
                        .format(
                            fun=event.name,
                            val=event.var.name)
                elif arity == 'InjectiveFunction':
                    given = 'GIVEN_INJECTIVE_FUNCTION {fun} {key} {val}'\
                        .format(
                            fun=event.name,
                            key=event.args[0],
                            val=event.var.name)
                elif arity == 'BinaryFunction':
                    given = 'GIVEN_BINARY_FUNCTION {fun} {lhs} {rhs} {val}'\
                        .format(
                            fun=event.name,
                            lhs=event.args[0],
                            rhs=event.args[1],
                            val=event.var.name)
                elif arity == 'SymmetricFunction':
                    given = 'GIVEN_SYMMETRIC_FUNCTION {fun} {lhs} {rhs} {val}'\
                        .format(
                            fun=event.name,
                            lhs=event.args[0],
                            rhs=event.args[1],
                            val=event.var.name)
                else:
                    raise ValueError('invalid arity: {}'.format(arity))
                header = [given]

                if diagonal:
                    header.append('IF_EQUAL {lhs} {rhs}'.format(
                        lhs=event.args[0],
                        rhs=event.args[1]))

                for cost, seq, plan in plans:
                    programs += [
                        '',
                        '# plan {}.{}: cost = {:0.1f}'.format(
                            group_id,
                            plan_id,
                            cost),
                        '# using {}'.format(sequent),
                        '# infer {}'.format(seq),
                    ]
                    programs += header
                    plan.program(programs)
                    plan_id += 1

            group_id += 1


def get_symbols_used_in(sequents, exprs):
    symbols = {}
    tokens = set()
    for seq in sequents:
        assert isinstance(seq, Sequent), seq
        for expr in seq.antecedents | seq.succedents:
            tokens |= set(expr.polish.split())
    for expr in exprs:
        assert isinstance(expr, Expression), expr
        tokens |= set(expr.polish.split())
    for token in list(tokens):
        if signature.get_arity(token) in signature.RELATION_ARITIES:
            try:
                tokens.add(try_negate_name(token))
            except NotNegatable:
                pass
    valid_arities = signature.FUNCTION_ARITIES | signature.RELATION_ARITIES
    for c in tokens:
        arity = signature.get_arity(c)
        if arity in valid_arities:
            symbols.setdefault(signature.get_arity(c), []).append(c)
    for val in symbols.itervalues():
        val.sort()
    return symbols


def write_programs(rules):
    sequents = set(rules)
    programs = []
    write_full_programs(programs, sequents)
    write_event_programs(programs, sequents)
    return programs


def write_symbols(rules, facts):
    sequents = set(rules) if rules else set()
    facts = set(facts) if facts else set()
    symbols = get_symbols_used_in(sequents, facts)
    symbols = [
        (arity, name)
        for arity, names in symbols.iteritems()
        if (arity in signature.FUNCTION_ARITIES or
            arity in signature.RELATION_ARITIES)
        for name in names
        if name != 'EQUAL'
    ]
    symbols.sort(key=lambda (arity, name): (signature.arity_sort(arity), name))
    return symbols



import re
from pomagma.compiler.compiler import compile_full
from pomagma.compiler.compiler import compile_given
from pomagma.compiler.compiler import get_events
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import Expression_0
from pomagma.compiler.expressions import Expression_1
from pomagma.compiler.expressions import Expression_2
from pomagma.compiler.plans import add_costs
from pomagma.compiler.sequents import Sequent


EQUAL = Expression_2('EQUAL')
LESS = Expression_2('LESS')
NLESS = Expression_2('NLESS')
CO = Expression_1('CO')
QUOTE = Expression_1('QUOTE')
APP = Expression_2('APP')
COMP = Expression_2('COMP')
JOIN = Expression_2('JOIN')
f = Expression_0('f')
g = Expression_0('g')
x = Expression_0('x')
y = Expression_0('y')
z = Expression_0('z')


outfile = None


def print_compiles(compiles):
    for cost, seq, plan in compiles:
        print '# cost = {0}'.format(cost)
        print '# infer {0}'.format(seq)
        print re.sub(': ', '\n', repr(plan))
        print


def _test_sequent(*args):
    sequent = Sequent(*args)
    print '-' * 78
    print 'Compiling full search: {0}'.format(sequent)
    compiles = compile_full(sequent)
    print_compiles(compiles)
    full_cost = add_costs(c for (c, _, _) in compiles)

    incremental_cost = None
    for event in get_events(sequent):
        print 'Compiling incremental search given: {0}'.format(event)
        compiles = compile_given(sequent, event)
        print_compiles(compiles)
        if event.args:
            cost = add_costs(c for (c, _, _) in compiles)
            if incremental_cost:
                incremental_cost = add_costs([incremental_cost, cost])
            else:
                incremental_cost = cost

    print '# full cost =', full_cost, 'incremental cost =', incremental_cost


def test_compile_I():
    I = Expression.make('I')
    _test_sequent(
        [],
        [EQUAL(APP(I, x), x)])


def test_compile_K():
    K = Expression.make('K')
    _test_sequent(
        [],
        [EQUAL(APP(APP(K, x), y), x)])


def test_compile_W():
    W = Expression.make('W')
    _test_sequent(
        [],
        [EQUAL(APP(APP(W, x), y), APP(APP(x, y), y))])


def test_compile_B_app():
    B = Expression.make('B')
    _test_sequent(
        [],
        [EQUAL(APP(APP(APP(B, x), y), z), APP(x, APP(y, z)))])


def test_compile_B_comp():
    B = Expression.make('B')
    _test_sequent(
        [],
        [EQUAL(APP(APP(B, x), y), COMP(x, y))])


def test_compile_app_comp():
    _test_sequent(
        [],
        [EQUAL(APP(COMP(x, y), z), APP(x, APP(y, z)))])


def test_compile_comp_assoc():
    _test_sequent(
        [],
        [EQUAL(COMP(COMP(x, y), z), COMP(x, COMP(y, z)))])


def test_compile_C():
    C = Expression.make('C')
    _test_sequent(
        [],
        [EQUAL(APP(APP(APP(C, x), y), z), APP(APP(x, z), y))])


def test_compile_S():
    # Compiling incremental search given: APP APP_x_y APP_y_z
    # cost = 2.70067540518
    # if S
    # for x let APP_S_x
    # for y if APP x y let APP_APP_S_x_y
    # for z if APP y z let APP_APP_APP_S_x_y_z
    # ensure EQUAL APP_APP_APP_S_x_y_z APP_APP_x_z_APP_y_z
    S = Expression.make('S')
    _test_sequent(
        [],
        [EQUAL(APP(APP(APP(S, x), y), z), APP(APP(x, z), APP(y, z)))])


def test_compile_Y():
    Y = Expression.make('Y')
    _test_sequent(
        [],
        [EQUAL(APP(Y, f), APP(f, APP(Y, f)))])


def test_compile_bot():
    BOT = Expression.make('BOT')
    _test_sequent(
        [],
        [LESS(BOT, x)])


def test_compile_reflexive():
    _test_sequent(
        [],
        [LESS(x, x)])


def test_compile_mono():
    _test_sequent(
        [LESS(x, y), LESS(f, g)],
        [LESS(APP(f, x), APP(g, y))])


def test_compile_mu():
    _test_sequent(
        [LESS(x, y)],
        [LESS(APP(f, x), APP(f, y))])


def test_compile_join_idem():
    _test_sequent(
        [],
        [EQUAL(JOIN(x, x), x)])


def test_compile_join_less():
    _test_sequent(
        [],
        [LESS(x, JOIN(x, y))])


def test_compile_co():
    # TODO use inverse iteration for injective_function::inverse_iterator
    # Compiling incremental search given: APP CO_x y
    # cost = 0.926425846749
    # for x if CO x let APP_y_x
    # ensure LESS APP_CO_x_y APP_y_x
    _test_sequent(
        [],
        [LESS(APP(CO(x), y), APP(y, x))])


def test_compile_comp_x_x_x():
    U = Expression.make('U')
    _test_sequent(
        [EQUAL(COMP(x, x), x)],
        [EQUAL(x, APP(U, x))])


def test_compile_eval():
    EVAL = Expression.make('EVAL')
    _test_sequent(
        [],
        [EQUAL(APP(EVAL, QUOTE(x)), x)])


def test_compile_qt_quote():
    QT = Expression.make('QT')
    _test_sequent(
        [],
        [EQUAL(APP(QT, QUOTE(x)), QUOTE(QUOTE(x)))])


def test_compile_ap_quote():
    AP = Expression.make('AP')
    _test_sequent(
        [],
        [EQUAL(APP(APP(AP, QUOTE(x)), QUOTE(y)), QUOTE(APP(x, y)))])



import sys
import inspect
import pomagma.util
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import Expression_1
from pomagma.compiler.expressions import Expression_2
from pomagma.compiler.plans import Ensure
from pomagma.compiler.plans import Iter
from pomagma.compiler.plans import IterInvBinary
from pomagma.compiler.plans import IterInvBinaryRange
from pomagma.compiler.plans import IterInvInjective
from pomagma.compiler.plans import Let
from pomagma.compiler.plans import Test
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.sequents import assert_normal
from pomagma.compiler.sequents import normalize
from pomagma.compiler.util import inputs
from pomagma.compiler.util import logger
from pomagma.compiler.util import memoize_args
from pomagma.compiler.util import set_with
from pomagma.compiler.util import set_without
from pomagma.compiler.util import sortedset
from pomagma.compiler.util import union


MIN_STACK_DEPTH = float('inf')


def stack_depth():
    global MIN_STACK_DEPTH
    depth = len(inspect.stack())
    MIN_STACK_DEPTH = min(MIN_STACK_DEPTH, depth)
    return depth - MIN_STACK_DEPTH


class DotPrinter(object):
    def __init__(self, out=sys.stdout):
        self.out = out
        self.count = 0

    def __call__(self, every=1000):
        assert every > 0
        self.count = (self.count + 1) % every
        if self.count == 0:
            self.out.write('.')
            self.out.flush()


print_dot = DotPrinter()


def POMAGMA_DEBUG_0(*args):
    pass


def POMAGMA_DEBUG_1(message, *args):
    print 'DEBUG{}'.format(' ' * stack_depth()),
    print message.format(*args)


if pomagma.util.LOG_LEVEL >= pomagma.util.LOG_LEVEL_DEBUG:
    POMAGMA_DEBUG = POMAGMA_DEBUG_1
else:
    POMAGMA_DEBUG = POMAGMA_DEBUG_0


EQUAL = Expression_2('EQUAL')
UNKNOWN = Expression_1('UNKNOWN')


@inputs(Sequent)
def compile_full(seq):
    results = []
    if seq.optional:
        logger('skipped optional rule {}', seq)
        return results
    for derived_seq in normalize(seq):
        context = frozenset()
        bound = frozenset()
        results.append(optimize_given(derived_seq, context, bound))
    assert results, 'failed to compile {0}'.format(seq)
    logger('derived {} rules from {}', len(results), seq)
    return results


@inputs(Sequent)
def get_events(seq):
    events = set()
    if seq.optional:
        logger('skipped optional rule {}', seq)
        return events
    free_vars = seq.vars
    for sequent in normalize(seq):
        for antecedent in sequent.antecedents:
            if antecedent.name == 'EQUAL':
                lhs, rhs = antecedent.args
                assert lhs.is_var() and rhs.is_var(), antecedent
                # HACK ignore equation antecedents
            else:
                events.add(antecedent)
        # HACK to deal with Equation args
        succedent = iter(sequent.succedents).next()
        for arg in succedent.args:
            if not arg.is_var():
                events.add(arg)
        antecedent_vars = union(a.vars for a in sequent.antecedents)
        for var in succedent.vars & free_vars - antecedent_vars:
            compound_count = sum(1 for arg in succedent.args if arg.args)
            in_count = sum(1 for arg in succedent.args if var in arg.vars)
            # if the var is in a compound in both succedent.args,
            # then succedent.args are sufficient events.
            if in_count < 2 or compound_count < 2:
                events.add(var)
    return events


def get_bound(atom):
    if atom.is_fun():
        return set_with(atom.vars, atom.var)
    else:
        return atom.vars


@inputs(Sequent, Expression)
def normalize_given(seq, atom, bound=None):
    if bound is None:
        bound = get_bound(atom)
    for normal in normalize(seq):
        if atom in normal.antecedents or atom.is_var():
            yield normal
        # HACK to deal with Equation args
        succedent = iter(normal.succedents).next()
        if succedent.name == 'EQUAL':
            lhs, rhs = succedent.args
            if lhs == atom:
                yield Sequent(
                    set_with(normal.antecedents, lhs),
                    set([EQUAL(lhs.var, rhs)]))
            elif rhs == atom:
                yield Sequent(
                    set_with(normal.antecedents, rhs),
                    set([EQUAL(lhs, rhs.var)]))


@inputs(Sequent, Expression)
def compile_given(seq, atom):
    context = frozenset([atom])
    bound = frozenset(get_bound(atom))
    normals = sorted(normalize_given(seq, atom, bound))
    assert normals, 'failed to compile {0} given {1}'.format(seq, atom)
    logger('derived {} rules from {} | {}', len(normals), atom, seq)
    return [optimize_given(n, context, bound) for n in normals]


@inputs(Sequent, frozenset, frozenset)
def optimize_given(seq, context, bound):
    assert_normal(seq)
    antecedents = sortedset(seq.antecedents - context)
    (succedent,) = list(seq.succedents)
    POMAGMA_DEBUG('{} | {} |- {}', list(bound), list(antecedents), succedent)
    plan = optimize_plan(antecedents, succedent, bound)
    plan.validate(bound)
    return plan.cost, seq, plan


@memoize_args
def optimize_plan(antecedents, succedent, bound):
    '''
    Iterate through the space of plans, narrowing heuristically.
    '''
    assert isinstance(antecedents, sortedset)
    assert isinstance(succedent.consts, sortedset)

    # ensure
    if not antecedents and succedent.vars <= bound:
        POMAGMA_DEBUG('ensure {}', succedent)
        return Ensure.make(succedent)

    # conditionals
    # HEURISTIC test eagerly in arbitrary order
    for a in antecedents:
        if a.is_rel():
            if a.vars <= bound:
                antecedents_a = sortedset(set_without(antecedents, a))
                POMAGMA_DEBUG('test relation {}', a)
                body = optimize_plan(antecedents_a, succedent, bound)
                return Test.make(a, body)
        else:
            assert a.is_fun(), a
            if a.vars <= bound and a.var in bound:
                antecedents_a = sortedset(set_without(antecedents, a))
                POMAGMA_DEBUG('test function {}', a)
                body = optimize_plan(antecedents_a, succedent, bound)
                return Test.make(a, body)

    # find & bind variable
    # HEURISTIC bind eagerly in arbitrary order
    for a in antecedents:
        if a.is_fun():
            if a.vars <= bound:
                assert a.var not in bound
                antecedents_a = sortedset(set_without(antecedents, a))
                bound_a = set_with(bound, a.var)
                POMAGMA_DEBUG('let {}', a)
                body = optimize_plan(antecedents_a, succedent, bound_a)
                return Let.make(a, body)
            else:
                # TODO find inverse if injective function
                pass

    results = []

    # iterate unknown
    if succedent.is_rel() and succedent.name != 'EQUAL':  # TODO handle EQUAL
        s_free = succedent.vars - bound
        if len(succedent.vars) == len(succedent.args) and len(s_free) == 1:
            v = iter(s_free).next()
            bound_v = set_with(bound, v)
            POMAGMA_DEBUG('iterate unknown {}', v)
            body = optimize_plan(antecedents, succedent, bound_v)
            results.append(Iter.make(v, Test.make(UNKNOWN(succedent), body)))

    # iterate forward
    forward_vars = set()
    for a in antecedents:
        a_free = a.vars - bound
        if len(a_free) == 1:
            forward_vars |= a_free
    for v in forward_vars:
        bound_v = set_with(bound, v)
        POMAGMA_DEBUG('iterate forward {}', v)
        body = optimize_plan(antecedents, succedent, bound_v)
        results.append(Iter.make(v, body))

    # iterate backward
    for a in antecedents:
        if a.is_fun() and a.args and a.var in bound and not (a.vars <= bound):
            nargs = len(a.args)
            a_free = a.vars - bound
            bound_v = bound | a_free
            antecedents_a = sortedset(set_without(antecedents, a))
            assert len(a_free) in [0, 1, 2]
            assert nargs in [0, 1, 2]
            POMAGMA_DEBUG('iterate backward {}', a)
            if nargs == 1 and len(a_free) == 1:
                # TODO injective function inverse need not be iterated
                body = optimize_plan(antecedents_a, succedent, bound_v)
                results.append(IterInvInjective.make(a, body))
            elif nargs == 2 and len(a_free) == 1 and len(a.vars) == 2:
                (fixed,) = list(a.vars - a_free)
                body = optimize_plan(antecedents_a, succedent, bound_v)
                results.append(IterInvBinaryRange.make(a, fixed, body))
            elif nargs == 2 and len(a_free) == 2:
                body = optimize_plan(antecedents_a, succedent, bound_v)
                results.append(IterInvBinary.make(a, body))

    # HEURISTIC iterate locally eagerly
    if results:
        return min(results)

    # iterate anything
    for v in union(a.vars for a in antecedents) | succedent.vars - bound:
        bound_v = set_with(bound, v)
        POMAGMA_DEBUG('iterate non-locally')
        body = optimize_plan(antecedents, succedent, bound_v)
        results.append(Iter.make(v, body))

    return min(results)



from nose.tools import assert_equal
from nose.tools import assert_set_equal
from pomagma.compiler.parser import parse_string_to_expr as parse
from pomagma.util.testing import for_each_kwargs


EXAMPLES = [
    {
        'polish': 'x',
        'polish_vars': set(['x']),
        'polish_consts': set(),
        'polish_terms': set(['x']),
    },
    {
        'polish': 'TOP',
        'polish_vars': set(),
        'polish_consts': set(['TOP']),
        'polish_terms': set(['TOP']),
    },
    {
        'polish': 'APP TOP x',
        'polish_vars': set(['x']),
        'polish_consts': set(['TOP']),
        'polish_terms': set(['x', 'TOP', 'APP TOP x']),
    },
    {
        'polish': 'APP x TOP',
        'polish_vars': set(['x']),
        'polish_consts': set(['TOP']),
        'polish_terms': set(['x', 'TOP', 'APP x TOP']),
    },
    {
        'polish': 'LESS x TOP',
        'polish_vars': set(['x']),
        'polish_consts': set(['TOP']),
        'polish_terms': set(['x', 'TOP']),
    },
    {
        'polish': 'LESS APP TOP x y',
        'polish_vars': set(['x', 'y']),
        'polish_consts': set(['TOP']),
        'polish_terms': set(['x', 'y', 'TOP', 'APP TOP x']),
    },
]

for example in EXAMPLES:
    example['expression'] = parse(example['polish'])
    example['vars'] = set(map(parse, example['polish_vars']))
    example['consts'] = set(map(parse, example['polish_consts']))
    example['terms'] = set(map(parse, example['polish_terms']))


@for_each_kwargs(EXAMPLES)
def test_polish(expression, polish, **unused):
    assert_equal(expression.polish, polish)


@for_each_kwargs(EXAMPLES)
def test_vars(expression, vars, **unused):
    assert_set_equal(expression.vars, vars)


@for_each_kwargs(EXAMPLES)
def test_consts(expression, consts, **unused):
    assert_set_equal(expression.consts, consts)


@for_each_kwargs(EXAMPLES)
def test_terms(expression, terms, **unused):
    assert_set_equal(expression.terms, terms)



import re
from nose.tools import assert_set_equal
from pomagma.analyst.compiler import compile_solver
from pomagma.util.testing import for_each_kwargs


SOLVE_EXAMPLES = [
    {
        'expr': 's',
        'theory': 'LESS APP V s s',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION V V_',
             'LETS_BINARY_FUNCTION_LHS APP V_ APP_V_s',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_POS_NEG s APP_V_s NRETURN_s',
             'LET_BINARY_FUNCTION APP V_ s APP_V_s',
             'IF_BINARY_RELATION NLESS APP_V_s s',
             'INFER_UNARY_RELATION NRETURN s'),
            ('FOR_NULLARY_FUNCTION V V_',
             'LETS_BINARY_FUNCTION_LHS APP V_ APP_V_s',
             'LETS_UNARY_RELATION RETURN RETURN_s',
             'FOR_POS_NEG s APP_V_s RETURN_s',
             'LET_BINARY_FUNCTION APP V_ s APP_V_s',
             'IF_BINARY_RELATION LESS APP_V_s s',
             'INFER_UNARY_RELATION RETURN s'),
        ]),
    },
    {
        'expr': 's',
        'theory': 'LESS APP s BOT BOT',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION BOT BOT_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_BOT BOT_',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_POS_NEG s APP_s_BOT NRETURN_s',
             'LET_BINARY_FUNCTION APP s BOT_ APP_s_BOT',
             'IF_BINARY_RELATION NLESS APP_s_BOT BOT_',
             'INFER_UNARY_RELATION NRETURN s'),
            ('FOR_NULLARY_FUNCTION BOT BOT_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_BOT BOT_',
             'LETS_UNARY_RELATION RETURN RETURN_s',
             'FOR_POS_NEG s APP_s_BOT RETURN_s',
             'LET_BINARY_FUNCTION APP s BOT_ APP_s_BOT',
             'IF_BINARY_RELATION LESS APP_s_BOT BOT_',
             'INFER_UNARY_RELATION RETURN s'),
        ]),
    },
    {
        'expr': 's',
        'theory': 'EQUAL APP s I I',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION I I_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_I I_',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_POS_NEG s APP_s_I NRETURN_s',
             'LET_BINARY_FUNCTION APP s I_ APP_s_I',
             'IF_BINARY_RELATION NLESS APP_s_I I_',
             'INFER_UNARY_RELATION NRETURN s'),
            ('FOR_NULLARY_FUNCTION I I_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_I I_',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_POS_NEG s APP_s_I NRETURN_s',
             'LET_BINARY_FUNCTION APP s I_ APP_s_I',
             'IF_BINARY_RELATION NLESS I_ APP_s_I',
             'INFER_UNARY_RELATION NRETURN s'),
            ('FOR_NULLARY_FUNCTION I I_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_I I_',
             'LETS_UNARY_RELATION RETURN RETURN_s',
             'FOR_POS_NEG s APP_s_I RETURN_s',
             'LET_BINARY_FUNCTION APP s I_ APP_s_I',
             'IF_EQUAL APP_s_I I_',
             'INFER_UNARY_RELATION RETURN s'),
        ]),
    },
    {
        'expr': 's',
        'theory': 'LESS TOP APP s TOP',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION TOP TOP_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_TOP TOP_',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_POS_NEG s APP_s_TOP NRETURN_s',
             'LET_BINARY_FUNCTION APP s TOP_ APP_s_TOP',
             'IF_BINARY_RELATION NLESS TOP_ APP_s_TOP',
             'INFER_UNARY_RELATION NRETURN s'),
            ('FOR_NULLARY_FUNCTION TOP TOP_',
             'LETS_BINARY_FUNCTION_RHS APP APP_s_TOP TOP_',
             'LETS_UNARY_RELATION RETURN RETURN_s',
             'FOR_POS_NEG s APP_s_TOP RETURN_s',
             'LET_BINARY_FUNCTION APP s TOP_ APP_s_TOP',
             'IF_BINARY_RELATION LESS TOP_ APP_s_TOP',
             'INFER_UNARY_RELATION RETURN s'),
        ]),
    },
    {
        'expr': 's',
        'theory': '''
            NLESS x BOT
            --------------
            LESS I APP s x
            ''',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION I I_',
             'FOR_NULLARY_FUNCTION BOT BOT_',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_NEG s NRETURN_s',
             'LETS_BINARY_RELATION_RHS NLESS NLESS_x_BOT BOT_',
             'LETS_BINARY_FUNCTION_LHS APP s APP_s_x',
             'FOR_POS_POS x NLESS_x_BOT APP_s_x',
             'LET_BINARY_FUNCTION APP s x APP_s_x',
             'IF_BINARY_RELATION NLESS I_ APP_s_x',
             'INFER_UNARY_RELATION NRETURN s'),
        ]),
    },
    {
        'expr': 's',
        'theory': '''
            NLESS x I
            ----------------
            LESS TOP APP s x
            ''',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION I I_',
             'FOR_NULLARY_FUNCTION TOP TOP_',
             'LETS_UNARY_RELATION NRETURN NRETURN_s',
             'FOR_NEG s NRETURN_s',
             'LETS_BINARY_RELATION_RHS NLESS NLESS_x_I I_',
             'LETS_BINARY_FUNCTION_LHS APP s APP_s_x',
             'FOR_POS_POS x NLESS_x_I APP_s_x',
             'LET_BINARY_FUNCTION APP s x APP_s_x',
             'IF_BINARY_RELATION NLESS TOP_ APP_s_x',
             'INFER_UNARY_RELATION NRETURN s'),
        ]),
    },
    {
        'expr': 's',
        'theory': '''
            # The entire theory of SEMI:
            LESS APP V s s       NLESS x BOT      NLESS x I
            LESS APP s BOT BOT   --------------   ----------------
            EQUAL APP s I I      LESS I APP s x   LESS TOP APP s x
            LESS TOP APP s TOP
            ''',
        'expected_programs': set([]),  # defined below
    },
    {
        'expr': 'x',
        'theory': '',
        'expected_programs': set([
            ('LETS_UNARY_RELATION RETURN RETURN_x',
             'FOR_NEG x RETURN_x',
             'INFER_UNARY_RELATION RETURN x'),
        ]),
    },
    {
        'expr': 'x',
        'theory': 'LESS x I',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION I I_',
             'LETS_BINARY_RELATION_RHS LESS LESS_x_I I_',
             'LETS_UNARY_RELATION RETURN RETURN_x',
             'FOR_POS_NEG x LESS_x_I RETURN_x',
             'INFER_UNARY_RELATION RETURN x'),
            ('FOR_NULLARY_FUNCTION I I_',
             'LETS_BINARY_RELATION_RHS NLESS NLESS_x_I I_',
             'LETS_UNARY_RELATION NRETURN NRETURN_x',
             'FOR_POS_NEG x NLESS_x_I NRETURN_x',
             'INFER_UNARY_RELATION NRETURN x'),
        ]),
    },
    {
        'expr': 'FUN f APP APP f s r',
        'theory': '''
            NLESS s BOT
            NLESS r BOT
            LESS COMP r s I
            ''',
        'expected_programs': set([
            ('FOR_NULLARY_FUNCTION C C_',
             'FOR_NULLARY_FUNCTION I I_',
             'FOR_NULLARY_FUNCTION BOT BOT_',
             'FOR_BINARY_FUNCTION_LHS_RHS APP C_ I_ APP_C_I',
             'LETS_BINARY_RELATION_RHS NLESS NLESS_s_BOT BOT_',
             'LETS_BINARY_FUNCTION_LHS APP APP_C_I APP_APP_C_I_s',
             'FOR_POS_POS s NLESS_s_BOT APP_APP_C_I_s',
             'LET_BINARY_FUNCTION APP APP_C_I s APP_APP_C_I_s',
             'FOR_BINARY_FUNCTION_LHS_RHS APP C_ APP_APP_C_I_s '
                'APP_C_APP_APP_C_I_s',
             'LETS_BINARY_RELATION_RHS NLESS NLESS_r_BOT BOT_',
             'LETS_BINARY_FUNCTION_RHS COMP COMP_r_s s',
             'LETS_BINARY_FUNCTION_LHS APP APP_C_APP_APP_C_I_s '
                'APP_APP_C_APP_APP_C_I_s_r',
             'FOR_POS_POS_POS r NLESS_r_BOT COMP_r_s '
                'APP_APP_C_APP_APP_C_I_s_r',
             'LET_BINARY_FUNCTION COMP r s COMP_r_s',
             'LET_BINARY_FUNCTION APP APP_C_APP_APP_C_I_s r '
                'APP_APP_C_APP_APP_C_I_s_r',
             'IF_BINARY_RELATION LESS COMP_r_s I_',
             'INFER_UNARY_RELATION RETURN APP_APP_C_APP_APP_C_I_s_r')
        ]),
    },
]

SOLVE_EXAMPLES[6]['expected_programs'] = set(
    p
    for e in SOLVE_EXAMPLES[:6]
    for p in e['expected_programs']
    if not re.search(' RETURN ', p[-1])
)


def parse_programs(script):
    script = re.sub('#.*\n', '', script)
    scripts = re.split('\n\n', script)
    return set(tuple(s.strip().splitlines()) for s in scripts)


@for_each_kwargs(SOLVE_EXAMPLES)
def test_compile_solver(expr, theory, expected_programs):
    script = compile_solver(expr, theory)
    actual_programs = parse_programs(script)
    assert_set_equal(expected_programs, actual_programs)



from pomagma.atlas.structure_pb2 import Structure
from pomagma.io import blobstore
from pomagma.io.protobuf import InFile
import contextlib
import os
import pomagma.cartographer
import pomagma.surveyor
import pomagma.theorist
import pomagma.util


@contextlib.contextmanager
def chdir(theory, init=False):
    path = os.path.join(pomagma.util.DATA, 'atlas', theory)
    if init:
        assert not os.path.exists(path), 'World map is already initialized'
        os.makedirs(path)
    else:
        assert os.path.exists(path), 'First initialize world map'
    with pomagma.util.chdir(path):
        yield


@contextlib.contextmanager
def load(theory, world, address=None, **opts):
    with pomagma.util.mutex(world):
        with pomagma.cartographer.load(theory, world, address, **opts) as db:
            yield db


def update_theory(theory, world, updated, dry_run=False, **opts):
    assert not os.path.exists(updated)
    with pomagma.util.mutex(world):
        assert os.path.exists(world)
        size = get_item_count(world)
        old_hash = get_hash(world)
        pomagma.surveyor.survey(theory, world, updated, size, **opts)
        new_hash = get_hash(updated)
        if new_hash == old_hash:
            print 'theory did not change'
            os.remove(updated)
            return False
        else:
            print 'theory changed'
            if dry_run:
                os.remove(updated)
            else:
                os.rename(updated, world)
            return True


def update_language(theory, init, world, updated, **opts):
    assert os.path.exists(init)
    assert not os.path.exists(updated)
    with pomagma.util.mutex(world):
        assert os.path.exists(world)
        with pomagma.cartographer.load(theory, init, **opts) as db:
            db.aggregate(world)
            db.validate()
            db.dump(updated)
        os.rename(updated, world)
    os.remove(init)


def update_format(theory, source, destin, **opts):
    print 'converting {} {} -> {}'.format(theory, source, destin)
    assert source != destin
    with pomagma.util.mutex(source):
        assert os.path.exists(source)
        assert not os.path.exists(destin)
        with pomagma.cartographer.load(theory, source, **opts) as db:
            db.validate()
            db.dump(destin)


def assume(theory, world, updated, theorems, **opts):
    assert not os.path.exists(updated)
    with pomagma.util.mutex(world):
        assert os.path.exists(world)
        with pomagma.cartographer.load(theory, world, **opts) as db:
            db.assume(theorems)
            db.validate()
            db.dump(updated)
        os.rename(updated, world)


def find(path):
    return filter(os.path.isfile, [
        os.path.abspath(os.path.join(root, filename))
        for root, dirnames, filenames in os.walk(path)
        for filename in filenames
    ])


def find_used_blobs(root):
    root = os.path.abspath(root)
    used_blobs = set()
    for filename in find(root):
        if filename.endswith('.pb'):
            with open(filename) as f:
                for line in f:
                    hexdigest = line.strip()
                    assert blobstore.RE_BLOB.match(hexdigest), hexdigest
                    used_blobs.add(hexdigest)
    return used_blobs


def garbage_collect(grace_period_days=blobstore.GRACE_PERIOD_DAYS):
    used_blobs = find_used_blobs(pomagma.util.DATA)
    blobstore.garbage_collect(used_blobs, grace_period_days)
    blobstore.validate_blobs()


def get_ext(filename):
    parts = filename.split('.')
    while parts[-1] in ['gz', 'bz2', '7z']:
        parts = parts[:-1]
    ext = parts[-1]
    assert ext in ['pb'], 'unsupported filetype: {}'.format(filename)
    return ext


def pb_load(filename):
    with InFile(blobstore.find_blob(blobstore.load_blob_ref(filename))) as f:
        structure = Structure()
        f.read(structure)
        return structure


def count_obs(structure):
    points = structure.getNode('/carrier/points')
    item_dim = max(points)
    item_count, = points.shape
    return item_dim, item_count


def get_hash(filename):
    assert get_ext(filename) == 'pb'
    return pb_load(filename).hash  # FIXME this is a string, not a list


def get_info(filename):
    assert get_ext(filename) == 'pb'
    item_count = pb_load(filename).carrier.item_count
    item_dim = item_count
    return {'item_dim': item_dim, 'item_count': item_count}


def get_item_count(filename):
    return get_info(filename)['item_count']


def get_filesize(filename):
    return os.stat(filename).st_size


def print_info(filename):
    assert get_ext(filename) == 'pb'
    files = [filename]
    files += map(blobstore.find_blob, blobstore.iter_blob_refs(filename))
    print 'file_count =', len(files)
    print 'byte_count =', sum(map(get_filesize, files))
    structure = pb_load(filename)
    print 'item_dim =', structure.carrier.item_count
    print 'item_count =', structure.carrier.item_count
    print structure



import functools


def for_each(examples):

    def decorator(fun):

        def fun_one(i):
            fun(examples[i])

        @functools.wraps(fun)
        def decorated():
            for i in xrange(len(examples)):
                yield fun_one, i

        return decorated
    return decorator


def for_each_kwargs(examples):

    def decorator(fun):

        def fun_one(i):
            fun(**examples[i])

        @functools.wraps(fun)
        def decorated():
            for i in xrange(len(examples)):
                yield fun_one, i

        return decorated
    return decorator


def for_each_context(get_context, examples):

    def decorator(fun):
        state = {}

        def fun_one(i):
            fun(state['context'], examples[i])

        @functools.wraps(fun)
        def decorated():
            with get_context() as context:
                state['context'] = context
                for i in xrange(len(examples)):
                    yield fun_one, i
            del state['context']

        return decorated
    return decorator



from itertools import izip
from nose import SkipTest
from nose.tools import assert_false
from pomagma.atlas.bootstrap import THEORY
from pomagma.atlas.bootstrap import WORLD
from pomagma.util.testing import for_each_context
import os
import pomagma.analyst
import pomagma.cartographer
import pomagma.surveyor
import pomagma.util
import simplejson as json

DATA = os.path.join(pomagma.util.DATA, 'test', 'debug', 'atlas', THEORY)
ADDRESS = 'ipc://{}'.format(os.path.join(DATA, 'socket'))
OPTIONS = {
    'log_file': os.path.join(DATA, 'analyst_test.log'),
    'log_level': pomagma.util.LOG_LEVEL_DEBUG,
}


def json_load(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    with open(filename) as f:
        return json.load(f)


SIMPLIFY_EXAMPLES = json_load('testdata/simplify_examples.json')
SIMPLIFY_EXAMPLES += json_load('testdata/simplify_sugar_examples.json')
VALIDATE_EXAMPLES = json_load('testdata/validate_examples.json')
CORPUS = json_load('testdata/corpus.json')


def setup_module():
    if not os.path.exists(WORLD):
        print 'Building test fixture', WORLD
        dirname = os.path.dirname(WORLD)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        min_size = pomagma.util.MIN_SIZES[THEORY]
        pomagma.surveyor.init(THEORY, WORLD, min_size)
        with pomagma.cartographer.load(THEORY, WORLD) as db:
            db.normalize()
            db.dump(WORLD)


def serve(address=ADDRESS):
    setup_module()
    return pomagma.analyst.serve(THEORY, WORLD, address, **OPTIONS)


def load():
    return pomagma.analyst.load(THEORY, WORLD, **OPTIONS)


def test_ping():
    print 'starting server'
    server = serve()
    try:
        print 'connecting client'
        with server.connect() as client:
            for _ in xrange(10):
                print 'pinging server'
                client.ping()
    finally:
        print 'stopping server'
        server.stop()


def test_ping_id():
    expected = 'test'
    with load() as db:
        actual = db.ping_id(expected)
    assert actual == expected


def test_inference():
    with load() as db:
        print 'Testing analyst inference'
        fail_count = db.test_inference()
    assert fail_count == 0, 'analyst failed with {} errors'.format(fail_count)


def assert_equal_example(expected, actual, example, cmp=cmp):
    assert_false(cmp(expected, actual), '\n'.join([
        'failed {}'.format(example),
        'expected: {}'.format(expected),
        'actual: {}'.format(actual)
    ]))


def assert_examples(examples, expected, actual, cmp=cmp):
    assert len(expected) == len(examples)
    assert len(actual) == len(examples)
    for example, e, a in izip(examples, expected, actual):
        print '{} : {}'.format(example, e)
        if cmp(e, a):
            print 'WARNING {}\n  expected: {}\n  actual: {}'.format(
                example, e, a)
    for example, e, a in izip(examples, expected, actual):
        assert_equal_example(e, a, example, cmp)


def cmp_trool(x, y):
    if x is None or y is None:
        return 0
    else:
        return cmp(x, y)


def cmp_validity(x, y):
    return (cmp_trool(x['is_top'], y['is_top']) or
            cmp_trool(x['is_bot'], y['is_bot']))


def transpose(lists):
    return map(list, izip(* lists))


def test_simplify():
    codes, expected = transpose(SIMPLIFY_EXAMPLES)
    with load() as db:
        actual = db.simplify(codes)
    assert_examples(codes, expected, actual)


SOLVE_EXAMPLES = [
    {
        'var': 'x',
        'theory': 'LESS x I   LESS I x',
        'necessary': ['I'],
    },
    {
        'var': 'x',
        'theory': 'LESS x I   LESS I x',
        'max_solutions': 999,
        'necessary': ['I'],
    },
    {
        'var': 'x',
        'theory': 'LESS TOP x',
        'necessary': ['TOP'],
    },
    {
        'var': 'x',
        'theory': 'FIXES TOP x',
        'necessary': ['TOP'],
    },
    {
        'var': 'x',
        'theory': 'EQUAL APP x x x',
        'max_solutions': 4,
        'necessary': ['I', 'BOT', 'TOP', 'V'],
    },
    {
        'var': 'x',
        'theory': 'EQUAL APP x x x   NLESS x BOT',
        'max_solutions': 3,
        'necessary': ['I', 'TOP', 'V'],
    },
    {
        'var': 'x',
        'theory': 'EQUAL APP x TOP APP x BOT   NLESS APP K APP x I x',
        'necessary': [],
    },
    {
        'var': 's',
        'theory': 'LESS APP V s s',
        'necessary': ['I', 'TOP', 'V', 'APP CB TOP', 'APP V C'],
        'possible': [],
    },
    {
        'var': 's',
        'theory': 'LESS APP s BOT BOT',
        'necessary': ['B', 'C', 'I', 'BOT', 'Y'],
        'possible': [],
    },
    {
        'var': 's',
        'theory': 'EQUAL APP s I I',
        'necessary': ['B', 'CB', 'I', 'V', 'COMP B B'],
        'possible': [],
    },
    {
        'var': 's',
        'theory': 'LESS TOP APP s TOP',
        'necessary': ['B', 'C', 'I', 'Y', 'TOP'],
        'possible': [],
    },
    {
        'var': 's',
        'theory': '''
            NLESS x BOT
            --------------
            LESS I APP s x
            ''',
        'necessary': [],
        'possible': ['TOP', 'J', 'V', 'P', 'APP C K'],
    },
    {  # FIXME
        'skip': 'spurriously fails with either JOIN B CB or JOIN CB B',
        'var': 's',
        'theory': '''
            NLESS x I
            ----------------
            LESS TOP APP s x
            ''',
        'necessary': [],
        'possible': [
            'TOP',
            'JOIN B CB',
            'APP C Y',
            'JOIN CI B',
            'RAND CI B',
        ],
    },
    {
        'var': 's',
        'theory': '''
            # The entire theory of SEMI:
            LESS APP V s s       NLESS x BOT      NLESS x I
            LESS APP s BOT BOT   --------------   ----------------
            EQUAL APP s I I      LESS I APP s x   LESS TOP APP s x
            LESS TOP APP s TOP
            ''',
        'necessary': [],
        'possible': [
            'APP C Y',
            'RAND CI B',
            'JOIN CI CB',
            'RAND CI CB',
            'RAND C CI',
        ],
    },
    {
        'skip': 'compiler fails',  # FIXME
        'var': 'x',
        'theory': 'EQUAL x I',
        'necessary': ['I'],
    },
    {
        'skip': 'compiler fails',  # FIXME
        'var': 'x',
        'theory': 'NLESS x x',
        'necessary': [],
        'possible': [],
    },
]


@for_each_context(load, SOLVE_EXAMPLES)
def test_solve(db, example):
    if 'skip' in example:
        raise SkipTest(example['skip'])
    max_solutions = example.get('max_solutions', 5)
    actual = db.solve(example['var'], example['theory'], max_solutions)
    for key in ['necessary', 'possible']:
        if key in example:
            assert_equal_example(example[key], actual[key], (example, key))


def test_validate():
    expected, codes = transpose(VALIDATE_EXAMPLES)
    with load() as db:
        actual = db.validate(codes)
    assert_examples(codes, expected, actual, cmp_validity)


def validate_corpus(lines, max_attempts=100):
    with load() as db:
        for attempt in xrange(1, 1 + max_attempts):
            print 'validating corpus, attempt', attempt
            results = db.validate_corpus(lines)
            if not any(result['pending'] for result in results):
                for validity in results:
                    del validity['pending']
                return results
    raise ValueError(
        'validate_corpus did not complete in {} attempts'.format(max_attempts))


def test_validate_corpus():
    expected, lines = transpose(CORPUS)
    actual = validate_corpus(lines)
    assert_examples(lines, expected, actual, cmp_validity)


VALIDATE_FACTS_EXAMPLES = [
    ([], True),
    (['LESS BOT BOT'], True),
    (['LESS TOP TOP'], True),
    (['LESS I I'], True),
    (['NLESS BOT BOT'], False),
    (['NLESS TOP TOP'], False),
    (['NLESS I I'], False),
    (['LESS TOP BOT'], False),
    (['EQUAL x x'], True),
    (['EQUAL x TOP', 'LESS TOP x'], True),
    (['EQUAL x TOP', 'LESS x BOT'], False),
    (['LESS TOP x', 'LESS x BOT'], False),
    (['LESS TOP x', 'LESS x y', 'LESS y BOT'], False),
    (['LESS TOP x', 'LESS x y', 'LESS y z', 'LESS z BOT'], False),
    (['LESS TOP w', 'LESS w x', 'LESS x y', 'LESS y z', 'LESS z BOT'], False),
    (['LESS I f', 'LESS I APP f f'], True),
    (['LESS I f', 'LESS I COMP f f'], True),
    (['LESS I f', 'LESS I JOIN f f'], True),
    (['LESS I f', 'LESS APP f f BOT'], False),
    (['LESS I f', 'LESS COMP f f BOT'], False),
    (['LESS I f', 'LESS JOIN f f BOT'], False),
    (['LESS I x', 'LESS APP x x y', 'LESS I COMP y y'], True),
    (['LESS I x', 'LESS APP x x y', 'NLESS I COMP y y'], False),
    (['LESS I x', 'NLESS y I', 'LESS APP x y z', 'LESS z I'], False),
    # TODO these demonstrate weakness in the solver
    (['NLESS x x'], True),
    (['LESS x y', 'NLESS y x'], True),
    (['LESS x y', 'LESS y z', 'NLESS z x'], True),
    # FIXME these fail and should pass
    (['LESS BOT I'], True),
    (['LESS I TOP'], True),
    (['NLESS TOP BOT'], True),
    (['NLESS I BOT'], True),
    (['NLESS TOP I'], True),
    (['LESS BOT TOP'], True),
    (['EQUAL x TOP', 'LESS BOT x'], True),
    (['LESS I x', 'NLESS y I', 'LESS APP x y z', 'NLESS z I'], True),
]


def validate_facts(db, facts, max_attempts=100):
    for attempt in xrange(1, 1 + max_attempts):
        print 'validating facts, attempt', attempt
        validity = db.validate_facts(facts)
        if validity is not None:
            return validity
    raise ValueError(
        'validate_facts did not complete in {} attempts'.format(max_attempts))


def test_validate_facts():
    with load() as db:
        facts, expected = transpose(VALIDATE_FACTS_EXAMPLES)
        actual = [validate_facts(db, f) for f in facts]
    try:
        assert_examples(facts, expected, actual, cmp_trool)
    except AssertionError as e:
        # raise
        raise SkipTest(e)


def test_get_histogram():
    lines = [line for _, line in CORPUS]
    with load() as db:
        for _ in xrange(10):
            db.validate_corpus(lines)
        histogram = db.get_histogram()
    print 'histogram:', histogram
    assert histogram['obs']
    assert histogram['symbols']


def validate_language(language):
    total = sum(language.itervalues())
    assert abs(total - 1) < 1e-4, 'bad total: {}'.format(total)
    assert isinstance(language, dict), language
    for key, val in language.iteritems():
        assert isinstance(key, str), key
        assert isinstance(val, float), val
        assert val > 0, '{} has no mass'.format(key)


def test_fit_language_histogram():
    with load() as db:
        lines = [line for _, line in CORPUS]
        for _ in xrange(10):
            db.validate_corpus(lines)
        histogram = db.get_histogram()
        language = db.fit_language(histogram)
        validate_language(language)


def test_fit_language_default():
    with load() as db:
        lines = [line for _, line in CORPUS]
        for _ in xrange(10):
            db.validate_corpus(lines)
        language = db.fit_language()
        validate_language(language)



import os
import glob
from parsable import parsable
parsable = parsable.Parsable()
import pomagma.util


DATA = pomagma.util.DATA


@parsable
def conjectures(theory):
    '''
    Write conjectures.html
    '''
    in_pattern = os.path.join(DATA, 'atlas', theory, '*conjectures*.facts')
    conjectures_files = list(glob.glob(in_pattern))
    assert conjectures_files, 'found no conjectures to report'
    conjectures_html = os.path.join(DATA, 'report', theory, 'conjectures.html')
    destin = os.path.dirname(conjectures_html)
    if not os.path.exists(destin):
        os.makedirs(destin)
    with open(conjectures_html, 'w') as html:
        html.write(
            '''
            <html>
            <head>
            <title> Pomagma {theory} Conjectures </title>
            </head>
            <body>
            '''.format(theory=theory)
        )
        for conjectures_file in conjectures_files:
            with open(conjectures_file) as conjectures:
                html.write('<p><pre>\n')
                html.write(conjectures.read())
                html.write('</pre></p>\n')
        html.write(
            '''
            </body>
            </html>
            '''
        )


if __name__ == '__main__':
    parsable()



import simplejson as json
from parsable import parsable
parsable = parsable.Parsable()
from pomagma.language.language_pb2 import Language, WeightedTerm


def json_load(filename):
    with open(filename) as f:
        return json.load(f)


def json_dump(data, filename):
    with open(filename, 'w') as f:
        json.dump(data, f, indent=4, sort_keys=True)


ARITY_TO_PB2 = {
    'NULLARY': WeightedTerm.NULLARY,
    'INJECTIVE': WeightedTerm.INJECTIVE,
    'BINARY': WeightedTerm.BINARY,
    'SYMMETRIC': WeightedTerm.SYMMETRIC,
}

ARITY_FROM_PB2 = {val: key for key, val in ARITY_TO_PB2.iteritems()}


def normalize_dict(grouped):
    '''
    L1-normalize the weights in a groupd dict in-place.
    '''
    total = sum(
        weight
        for group in grouped.itervalues()
        for weight in group.itervalues()
    )
    assert total > 0, "total weight is zero"
    scale = 1.0 / total

    for group in grouped.itervalues():
        for key in group:
            group[key] *= scale


def dict_to_language(grouped):
    '''
    Convert from grouped dict to protobuf format.

    Example Input:

        grouped = {
            'NULLARY': {
                'B': 1.0,
                'C': 1.30428,
                'I': 2.21841,
                'K': 2.6654,
                'S': 2.69459,
                },
            'BINARY': {
                'APP': 0.4,
                'COMP': 0.2,
                },
            #'INJECTIVE': {},  # empty, may be omitted
            #'SYMMETRIC': {},  # empty, may be omitted
            }
    '''
    grouped = grouped.copy()
    normalize_dict(grouped)
    language = Language()
    for arity, group in grouped.iteritems():
        for name, weight in group.iteritems():
            term = language.terms.add()
            term.name = name
            term.arity = ARITY_TO_PB2[arity.upper()]
            term.weight = weight

    return language


def language_to_dict(language):
    '''
    Convert from protobuf format to grouped dict.
    '''
    grouped = {}
    for term in language.terms:
        arity = ARITY_FROM_PB2[term.arity]
        grouped.setdefault(arity, {})[term.name] = term.weight
    return grouped


@parsable
def compile(json_in, language_out):
    '''
    Convert language from json to protobuf format.
    '''
    with open(json_in) as f:
        grouped = json.load(f)
    language = dict_to_language(grouped)
    with open(language_out, 'wb') as f:
        f.write(language.SerializeToString())


if __name__ == '__main__':
    parsable()



import os
import pomagma.util
import pomagma.theorist.diverge


BIN = os.path.join(pomagma.util.BIN, 'theorist')


def _count_facts(filename):
    count = 0
    if os.path.exists(filename):
        with open(filename) as f:
            for line in f:
                if line.split('#', 1)[0].strip():
                    count += 1
    return count


def try_prove_nless(
        theory,
        world_in,
        conjectures_in,
        conjectures_out,
        theorems_out,
        **opts):
    prev_theorem_count = _count_facts(theorems_out)
    pomagma.util.log_call(
        os.path.join(BIN, 'try_prove_nless'),
        pomagma.util.abspath(world_in),
        os.path.join(pomagma.util.LANGUAGE, '{}.language'.format(theory)),
        pomagma.util.abspath(conjectures_in),
        pomagma.util.abspath(conjectures_out),
        pomagma.util.abspath(theorems_out),
        **opts)
    theorem_count = _count_facts(theorems_out) - prev_theorem_count
    return theorem_count


def try_prove_diverge(conjectures_in, conjectures_out, theorems_out, **opts):
    theorem_count = pomagma.theorist.diverge.try_prove_diverge(
        pomagma.util.abspath(conjectures_in),
        pomagma.util.abspath(conjectures_out),
        pomagma.util.abspath(theorems_out),
        **opts)
    return theorem_count



import os
import pomagma.analyst
import pomagma.language.util
import pomagma.util

JSON = os.path.join(pomagma.util.LANGUAGE, '{}.json')
PROTO = os.path.join(pomagma.util.LANGUAGE, '{}.language')


def fit_language(
        theory,
        address=pomagma.analyst.ADDRESS,
        log_file=None,
        log_level=0):

    language_json = JSON.format(theory)
    language_proto = PROTO.format(theory)

    def log_print(message, level):
        if level >= log_level:
            if log_file:
                pomagma.util.log_print(message, log_file)
            else:
                print message

    log_print('fitting language', pomagma.util.LOG_LEVEL_INFO)
    with pomagma.analyst.connect(address) as db:
        new_weights = db.fit_language()

    log_print('converting language', pomagma.util.LOG_LEVEL_DEBUG)
    language = pomagma.language.util.json_load(language_json)
    new_terms = sorted(new_weights.iterkeys())
    old_terms = sorted(key for group in language.itervalues() for key in group)
    assert new_terms == old_terms, '\n  '.join([
        'language mismatch,'
        'expected: {}'.format(old_terms),
        'actual: {}'.format(new_terms),
    ])
    for group in language:
        for key in group:
            group[key] = new_weights[key]

    log_print('writing {}'.format(language_json), pomagma.util.LOG_LEVEL_INFO)
    pomagma.language.util.json_dump(language, language_json)

    log_print('writing {}'.format(language_proto), pomagma.util.LOG_LEVEL_INFO)
    pomagma.language.util.compile(language_json, language_proto)



from parsable import parsable
import pomagma.analyst
import pomagma.analyst.compiler
import sys

theories = {
    'unit': '''
        CLOSED t         -----------------   LESS APP f TOP APP g TOP
        FIXES V t        EQUAL APP t x TOP   LESS APP f I APP g I
        FIXES t I        EQUAL APP t x I     ------------------------
        LESS APP J I t                       LESS COMP f t COMP g t
        ''',
    'semi': '''
        CLOSED t      -----------------   LESS APP f TOP APP g TOP
        FIXES V t     EQUAL APP t x TOP   LESS APP f I APP g I
        FIXES t BOT   EQUAL APP t x I     LESS APP f BOT APP g BOT
        FIXES t I     EQUAL APP t x BOT   ------------------------
                                          LESS COMP f t COMP g t
        ''',
    'bool': '''
        CLOSED t      -----------------   LESS APP f TOP APP g TOP
        FIXES V t     EQUAL APP t x TOP   LESS APP f K APP g K
        FIXES t BOT   EQUAL APP t x K     LESS APP f F APP g F
        FIXES t K     EQUAL APP t x F     LESS APP f BOT APP g BOT
        FIXES t F     EQUAL APP t x BOT   ------------------------
                                          LESS COMP f t COMP g t
        ''',
    'boool': '''
        CLOSED t      -----------------   LESS APP f TOP APP g TOP
        FIXES V t     EQUAL APP t x TOP   LESS APP f J APP g J
        FIXES t BOT   EQUAL APP t x J     LESS APP f K APP g K
        FIXES t K     EQUAL APP t x K     LESS APP f F APP g F
        FIXES t F     EQUAL APP t x F     LESS APP f BOT APP g BOT
        FIXES t J     EQUAL APP t x BOT   ------------------------
                                          LESS COMP f t COMP g t
        ''',
    'unit_test': '''
                         -----------------   LESS APP f TOP APP g TOP
        FIXES V t        EQUAL APP t x TOP   LESS APP f I APP g I
        FIXES t I        EQUAL APP t x I     ------------------------
        LESS APP J I t                       LESS COMP f t COMP g t
        ''',
    'semi_test': '''
                      -----------------   LESS APP f TOP APP g TOP
        FIXES V t     EQUAL APP t x TOP   LESS APP f I APP g I
        FIXES t BOT   EQUAL APP t x I     LESS APP f BOT APP g BOT
        FIXES t I     EQUAL APP t x BOT   ------------------------
                                          LESS COMP f t COMP g t
        ''',
}


def print_solutions(solutions):
    print 'Necessary:'
    for term in solutions['necessary']:
        print '  {}'.format(term)
    print 'Possible:'
    for term in solutions['possible']:
        print '  {}'.format(term)


@parsable
def show_define(name=None):
    '''
    Print program generated by analyst.compiler.compile_solver used in define.
    Available types: unit, semi, bool, boool
    '''
    if name not in theories:
        print 'Try one of: {}'.format(' '.join(sorted(theories)))
        sys.exit(1)
    print pomagma.analyst.compiler.compile_solver('t', theories[name])


@parsable
def define(name=None, max_solutions=32, address=pomagma.analyst.ADDRESS):
    '''
    Conjecture definitions of a type.
    Available types: unit, semi, bool, boool
    '''
    if name not in theories:
        print 'Try one of: {}'.format(' '.join(sorted(theories)))
        sys.exit(1)
    with pomagma.analyst.connect(address) as db:
        solutions = db.solve('t', theories[name], max_solutions)
    print_solutions(solutions)
    return solutions


@parsable
def sr_pairs(max_solutions=32, address=pomagma.analyst.ADDRESS):
    '''
    Find section,retract pairs (i.e. pairs below A).
    '''
    expr = 'FUN f APP APP f s r'    # coding of the pair (s, r)
    theory = '''
        CLOSED s   NLESS s BOT      # s is closed and nontrivial
        CLOSED r   NLESS r BOT      # r is closed and nontrivial
        LESS COMP r s I             # (s, r) is a section-retract pair
        '''
    with pomagma.analyst.connect(address) as db:
        solutions = db.solve(expr, theory, max_solutions)
        pairs = (
            [(sr, True) for sr in solutions['necessary']] +
            [(sr, False) for sr in solutions['possible']])
        sections = db.simplify(['APP {} K'.format(sr) for sr, _ in pairs])
        retracts = db.simplify(['APP {} F'.format(sr) for sr, _ in pairs])
    parts = zip(pairs, sections, retracts)
    solutions = {
        'necessary': [(sr, s, r) for (sr, n), s, r in parts if n],
        'possible': [(sr, s, r) for (sr, n), s, r in parts if not n],
    }
    print_solutions(solutions)
    return solutions


if __name__ == '__main__':
    parsable()



import os
import pomagma.util
import pomagma.workers
from pomagma.util import DB


def test_structure_queue():
    with pomagma.util.in_temp_dir():
        queue = pomagma.workers.FileQueue('test.queue')
        assert not queue.get()
        test_file = DB('test')
        with open(test_file, 'w') as f:
            f.write('test')
        assert os.path.exists(test_file)
        queue.push(test_file)
        assert not os.path.exists(test_file)
        assert len(queue) == 1
        assert queue.try_pop(test_file)
        assert os.path.exists(test_file)
        assert len(queue) == 0



from pomagma.atlas.bootstrap import THEORY
from pomagma.atlas.bootstrap import WORLD
import os
import pomagma.cartographer
import pomagma.util

DATA = os.path.join(pomagma.util.DATA, 'test', 'debug', 'atlas', THEORY)
ADDRESS = 'ipc://{}'.format(os.path.join(DATA, 'socket'))
OPTIONS = {
    'log_file': os.path.join(DATA, 'cartographer_test.log'),
    'log_level': pomagma.util.LOG_LEVEL_DEBUG,
}


def test_ping():
    print 'starting server'
    server = pomagma.cartographer.serve(THEORY, WORLD, ADDRESS, **OPTIONS)
    try:
        print 'connecting client'
        with server.connect() as client:
            for _ in xrange(10):
                print 'pinging server'
                client.ping()
    finally:
        print 'stopping server'
        server.stop()



from contextlib2 import ExitStack
from pomagma import analyst
from pomagma import atlas
from pomagma import cartographer
from pomagma import surveyor
from pomagma import theorist
from pomagma.util import DB
from pomagma.io import blobstore
import mock
import os
from parsable import parsable
import pomagma.util

PROFILERS = {
    'time': '/usr/bin/time --verbose',
    'valgrind': 'valgrind --leak-check=full --track-origins=yes',
    'cachegrind': 'valgrind --tool=cachegrind',
    'callgrind':
        'valgrind --tool=callgrind --callgrind-out-file=callgrind.out',
    'helgrind': 'valgrind --tool=helgrind --read-var-info=yes',
}


def _test_atlas(theory):
    '''
    Test basic operations in one theory:
        init, validate, copy, survey, aggregate,
        conjecture_diverge, conjecture_equal
    '''
    buildtype = 'debug' if pomagma.util.debug else 'release'
    path = os.path.join(pomagma.util.DATA, 'test', buildtype, 'atlas', theory)
    if os.path.exists(path):
        os.system('rm -f {}/*'.format(path))
    else:
        os.makedirs(path)
    with ExitStack() as stack:
        with_ = stack.enter_context
        with_(pomagma.util.chdir(path))
        with_(mock.patch('pomagma.util.BLOB_DIR', new=path))
        with_(pomagma.util.mutex(block=False))

        min_size = pomagma.util.MIN_SIZES[theory]
        dsize = min(64, 1 + min_size)
        sizes = [min_size + i * dsize for i in range(10)]
        opts = {'log_file': 'test.log', 'log_level': 2}
        diverge_conjectures = 'diverge_conjectures.facts'
        diverge_theorems = 'diverge_theorems.facts'
        equal_conjectures = 'equal_conjectures.facts'
        equal_theorems = 'equal_theorems.facts'

        surveyor.init(theory, DB(0), sizes[0], **opts)
        changed = atlas.update_theory(theory, DB(0), DB(1), **opts)
        assert not changed
        with cartographer.load(theory, DB(0), **opts) as db:
            db.validate()
            db.dump(DB(1))
            expected_size = atlas.get_item_count(DB(1))
            actual_size = db.info()['item_count']
            assert actual_size == expected_size
        surveyor.survey(theory, DB(1), DB(2), sizes[1], **opts)
        with cartographer.load(theory, DB(2), **opts) as db:
            db.trim([{'size': sizes[0], 'filename': DB(3)}])
        surveyor.survey(theory, DB(3), DB(4), sizes[1], **opts)
        with cartographer.load(theory, DB(2), **opts) as db:
            db.aggregate(DB(4))
            db.dump(DB(5))
        with cartographer.load(theory, DB(5), **opts) as db:
            db.aggregate(DB(0))
            db.dump(DB(6))
            digest5 = atlas.get_hash(DB(5))
            digest6 = atlas.get_hash(DB(6))
            assert digest5 == digest6

            counts = db.conjecture(diverge_conjectures, equal_conjectures)
            assert counts['diverge_count'] > 0, counts['diverge_count']
            assert counts['equal_count'] > 0, counts['equal_count']
            if theory != 'h4':
                theorem_count = theorist.try_prove_diverge(
                    diverge_conjectures,
                    diverge_conjectures,
                    diverge_theorems,
                    **opts)
                assert theorem_count > 0, theorem_count
                counts = db.assume(diverge_theorems)
                assert counts['pos'] + counts['neg'] > 0, counts
                assert counts['ignored'] == 0, counts
                db.validate()
                db.dump(DB(6))
        theorem_count = theorist.try_prove_nless(
            theory,
            DB(6),
            equal_conjectures,
            equal_conjectures,
            equal_theorems,
            **opts)
        # assert theorem_count > 0, theorem_count

        if theory != 'h4':
            with cartographer.load(theory, DB(6), **opts) as db:
                db.assume(equal_theorems)
                if theorem_count > 0:
                    assert counts['merge'] > 0, counts
                db.validate()
                for priority in [0, 1]:
                    while db.infer(priority):
                        db.validate()
                for priority in [0, 1]:
                    assert not db.infer(priority)
                db.dump(DB(7))
            with analyst.load(theory, DB(7), **opts) as db:
                fail_count = db.test_inference()
                assert fail_count == 0, 'analyst.test_inference failed'
        blobstore.validate_blobs()


@parsable
def test_atlas(theory='all'):
    '''
    Test atlas exploration and analysis operations.
    '''
    if theory == 'all':
        theories = pomagma.util.MIN_SIZES.keys()
        theories.sort(key=pomagma.util.MIN_SIZES.__getitem__)
    else:
        theories = [theory]

    for theory in theories:
        print '-' * 78
        print 'Testing', theory
        _test_atlas(theory)


@parsable
def test_analyst(theory):
    '''
    Test analyst approximation on normalized world map.
    '''
    opts = {'log_file': 'test.log', 'log_level': 2}
    with atlas.chdir(theory):
        world = DB('world.normal')
        assert os.path.exists(world), 'First initialize normalized world'
        with analyst.load(theory, world, **opts) as db:
            fail_count = db.test()
    assert fail_count == 0, 'Failed {} cases'.format(fail_count)
    print 'Passed analyst test'


@parsable
def profile_misc():
    '''
    Profile misc libraries.
    '''
    buildtype = 'debug' if pomagma.util.debug else 'release'
    log_file = os.path.join(pomagma.util.DATA, 'profile', buildtype + '.log')
    if os.path.exists(log_file):
        os.remove(log_file)
    opts = {'log_file': log_file, 'log_level': 2}
    cmds = [
        os.path.abspath(os.path.join(root, filename))
        for root, dirnames, filenames in os.walk(pomagma.util.BIN)
        for filename in filenames
        if filename.endswith('_profile')
    ]
    assert cmds, 'no profiles found in {}'.format(pomagma.util.BIN)
    for cmd in cmds:
        print 'Profiling', cmd
        pomagma.util.log_call(cmd, **opts)
    pomagma.util.check_call('cat', log_file)


@parsable
def profile_surveyor(theory='skj', grow_by=64, extra_size=0, tool='time'):
    '''
    Profile surveyor on random region of world.
    Available tools: time, valgrind, cachegrind, callgrind, helgrind
    '''
    size = pomagma.util.MIN_SIZES[theory] + extra_size
    with atlas.chdir(theory):
        opts = {'log_file': 'profile.log', 'log_level': 2}
        region = DB('region.{:d}'.format(size))
        temp = pomagma.util.temp_name(DB('profile'))
        world = DB('world')
        if not os.path.exists(region):
            print 'Creating {} for profile'.format(region)
            assert os.path.exists(world), 'First initialize world map'
            with cartographer.load(theory, world, **opts) as db:
                db.trim([{'size': size, 'filename': region}])
        opts.setdefault('runner', PROFILERS.get(tool, tool))
        surveyor.survey(theory, region, temp, size + grow_by, **opts)
        os.remove(temp)


@parsable
def profile_cartographer(theory='skj', extra_size=0, tool='time', infer=True):
    '''
    Profile cartographer load-infer-dump on region of world.
    Available tools: time, valgrind, cachegrind, callgrind, helgrind
    '''
    size = pomagma.util.MIN_SIZES[theory] + extra_size
    with atlas.chdir(theory):
        opts = {'log_file': 'profile.log', 'log_level': 2}
        region = DB('region.{:d}'.format(size))
        temp = pomagma.util.temp_name(DB('profile'))
        world = DB('world')
        world_size = atlas.get_item_count(world)
        if size >= world_size:
            print 'Using world of size {}'.format(world_size)
            region = world
        elif not os.path.exists(region):
            print 'Creating {} for profile'.format(region)
            assert os.path.exists(world), 'First initialize world map'
            with cartographer.load(theory, world, **opts) as db:
                db.trim([{'size': size, 'filename': region}])
        opts.setdefault('runner', PROFILERS.get(tool, tool))
        with cartographer.load(theory, region, **opts) as db:
            if infer:
                for priority in [0, 1]:
                    count = db.infer(priority)
                    print 'Proved {} theorems'.format(count)
            db.dump(temp)
        print 'Hash:', atlas.get_hash(temp)
        os.remove(temp)


@parsable
def coverity():
    '''
    Check pomagma build with coverity. (see http://coverity.com)
    '''
    pomagma.util.coverity()


if __name__ == '__main__':
    parsable()



import contextlib
import errno
import fcntl
import itertools
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import timeit
import uuid

SRC = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
ROOT = os.path.dirname(SRC)
THEORY = os.path.join(SRC, 'theory')
LANGUAGE = os.path.join(SRC, 'language')
DATA = os.path.join(ROOT, 'data')
DB_FORMAT = os.environ.get('POMAGMA_DB_FORMAT', 'pb')
assert DB_FORMAT in ['pb'], DB_FORMAT
DB = '{{}}.{}'.format(DB_FORMAT).format
BLOB_DIR = os.path.join(DATA, 'blob')
debug = int(os.environ.get('POMAGMA_DEBUG', 0))
if debug:
    print 'Running in debug mode'
    BUILD = os.path.join(ROOT, 'build', 'debug')
else:
    BUILD = os.path.join(ROOT, 'build', 'release')
CLEANUP_ON_ERROR = int(os.environ.get('CLEANUP_ON_ERROR', 1))
COVERITY = os.path.join(ROOT, 'cov-int')
BIN = os.path.join(BUILD, 'src')
TRAVIS_CI = 'TRAVIS' in os.environ and 'CI' in os.environ

LOG_LEVEL = int(os.environ.get('POMAGMA_LOG_LEVEL', 0))
LOG_LEVEL_ERROR = 0
LOG_LEVEL_WARNING = 1
LOG_LEVEL_INFO = 2
LOG_LEVEL_DEBUG = 3

MIN_SIZES = {
    'h4': 127,
    'sk': 1023,
    'skj': 1535,
    'skja': 2047,
    'skrj': 2047,
}


def optimal_db_sizes():
    '''Indefinitely iterate through optimal db sizes'''
    yield 512 - 1
    for i in itertools.count():
        yield 2 * 2 ** i * 512 - 1
        yield 3 * 2 ** i * 512 - 1


def suggest_region_sizes(min_size, max_size):
    '''Return set of optimal db sizes in [min_size, max_size]'''
    sizes = []
    for size in optimal_db_sizes():
        if size > max_size:
            return sizes
        if size >= min_size:
            sizes.append(size)


def TODO(message=''):
    raise NotImplementedError('TODO {}'.format(message))


def on_signal(sig):
    def decorator(handler):
        signal.signal(sig, handler)
        return handler
    return decorator


# adapted from
# http://blog.devork.be/2009/07/how-to-bring-running-python-program.html
@on_signal(signal.SIGUSR1)
def handle_pdb(sig, frame):
    import pdb
    pdb.Pdb().set_trace(frame)


def get_rss(pid):
    try:
        return int(subprocess.check_output(['ps', '-o', 'rss=', str(pid)]))
    except subprocess.CalledProcessError:
        return 0


def print_dot(out=sys.stdout):
    out.write('.')
    out.flush()


def random_uuid():
    return str(uuid.uuid4())


@contextlib.contextmanager
def chdir(path):
    old_path = os.path.abspath(os.path.curdir)
    try:
        # os.makedirs(path)
        print '# cd {}\n'.format(path)
        os.chdir(path)
        yield os.path.curdir
    finally:
        print '# cd {}\n'.format(old_path)
        os.chdir(old_path)


def temp_name(path):
    dirname, filename = os.path.split(path)
    # assert not filename.startswith('temp.'), path
    return os.path.join(dirname, 'temp.{}.{}'.format(os.getpid(), filename))


@contextlib.contextmanager
def temp_copy(path):
    dirname = os.path.dirname(path)
    assert not dirname or os.path.exists(dirname), path
    temp = temp_name(path)
    if os.path.exists(temp):
        shutil.rmtree(temp)
    yield temp
    if os.path.exists(temp):
        os.rename(temp, path)
    elif os.path.exists(path):
        os.remove(path)


@contextlib.contextmanager
def temp_copies(paths):
    temps = []
    for path in paths:
        dirname = os.path.dirname(path)
        assert not dirname or os.path.exists(dirname), path
        temp = temp_name(path)
        if os.path.exists(temp):
            shutil.rmtree(temp)
        temps.append(temp)
    yield temps
    for temp, path in zip(temps, paths):
        if os.path.exists(temp):
            os.rename(temp, path)
        elif os.path.exists(path):
            os.remove(path)


@contextlib.contextmanager
def in_temp_dir():
    path = os.path.abspath(tempfile.mkdtemp())
    cleanup = CLEANUP_ON_ERROR
    try:
        with chdir(path):
            yield path
        cleanup = True
    finally:
        if cleanup:
            print '# rm -rf {}\n'.format(path)
            shutil.rmtree(path)


class MutexLockedException(Exception):

    def __init__(self, filename):
        self.filename = os.path.abspath(filename)

    def __str__(self):
        return 'Failed to acquire lock on {}'.format(self.filename)


# Adapted from:
# http://blog.vmfarms.com/2011/03/cross-process-locking-and.html
@contextlib.contextmanager
def mutex(filename=None, block=True):
    if filename is None:
        mutex_filename = 'mutex'
    elif os.path.isdir(filename):
        mutex_filename = os.path.join(filename, 'mutex')
    else:
        mutex_filename = '{}.mutex'.format(filename)
    with open(mutex_filename, 'w') as fd:
        if block:
            try:
                fcntl.flock(fd, fcntl.LOCK_EX)
                fd.write('{}\n'.format(os.getpid()))
                fd.flush()
                yield
            finally:
                fcntl.flock(fd, fcntl.LOCK_UN)
        else:
            try:
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                fd.write('{}\n'.format(os.getpid()))
                fd.flush()
            except IOError, e:
                assert e.errno in [errno.EACCES, errno.EAGAIN]
                raise MutexLockedException(mutex_filename)
            else:
                try:
                    yield
                finally:
                    fcntl.flock(fd, fcntl.LOCK_UN)


@contextlib.contextmanager
def log_duration():
    try:
        start_time = timeit.default_timer()
        yield
    finally:
        duration = timeit.default_timer() - start_time
        print '# took {:0.3g} sec'.format(duration)


def abspath(path):
    return os.path.abspath(os.path.expanduser(path))


def ensure_abspath(filename):
    filename = abspath(filename)
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    return filename


def get_log_file(options):
    log_file = os.path.join(DATA, 'default.log')
    log_file = abspath(options.get('log_file', log_file))
    return log_file


def log_print(message, log_file):
    with open(log_file, 'a') as log:
        log.write(message)
        log.write('\n')
    print message


def make_env(options):
    options = dict(options)
    log_file = get_log_file(options)
    options['log_file'] = log_file
    log_dir = os.path.dirname(log_file)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    default_log_level = LOG_LEVEL_DEBUG if debug else LOG_LEVEL_INFO
    options.setdefault('log_level', default_log_level)
    options['blob_dir'] = os.path.abspath(options.get('blob_dir', BLOB_DIR))
    env = {
        'POMAGMA_{}'.format(key.upper()): str(val)
        for key, val in options.iteritems()
    }
    return env


def print_command(args, env={}):
    lines = ['{}={}'.format(key, val) for key, val in env.iteritems()]
    lines += args
    message = '{}\n'.format(' \\\n  '.join(lines))
    sys.stderr.write(message)


def check_call(*args):
    print_command(args)
    with log_duration():
        proc = subprocess.Popen(args)
        try:
            info = proc.wait()
        finally:
            if proc.poll() is None:
                proc.terminate()
    if info:
        sys.stderr.write('ERROR in {}'.format(' '.join(args)))
        sys.exit(info)


def print_logged_error(log_file):
    print
    print '==== LOG FILE ===='
    grep = ' '.join([
        'grep',
        '--before-context=40',
        '--after-context=3',
        '--ignore-case',
        '--color=always',
        '--text',
        '--max-count=1',
        'error'
    ])
    revgrep = 'tac {} | {} | tac'.format(log_file, grep)
    subprocess.call(revgrep, shell=True)


def get_stack_trace(binary):
    trace = '==== STACK TRACE ====\n'
    try:
        trace += subprocess.check_output([
            'gdb',
            binary,
            'core',
            '--batch',
            '-ex',
            'thread apply all bt',
        ])
    except subprocess.CalledProcessError:
        trace += 'ERROR stack trace failed'
    return trace


def prepare_core_dump():
    if os.path.exists('core'):
        os.remove('core')
    if subprocess.check_output('ulimit -c', shell=True).strip() == '0':
        print 'WARNING cannot write core file; try `ulimit -c unlimited`'


def log_call(*args, **options):
    '''
    Pass arguments to command line.
    Pass options into environment variables.
    Log process; if it crashes, dump stack trace to log file.
    '''
    args = map(str, args)
    args = options.pop('runner', '').split() + args
    extra_env = make_env(options)
    log_file = extra_env['POMAGMA_LOG_FILE']
    prepare_core_dump()
    print_command(args, extra_env)
    env = os.environ.copy()
    env.update(extra_env)
    with log_duration():
        proc = subprocess.Popen(args, env=env)
        try:
            info = proc.wait()
        finally:
            if proc.poll() is None:
                proc.terminate()
    if info:
        print_logged_error(log_file)
        trace = get_stack_trace(args[0])
        log_print(trace, log_file)
        sys.exit(info)


def log_Popen(*args, **options):
    '''
    Pass arguments to command line.
    Pass options into environment variables.
    Log process.
    '''
    args = map(str, args)
    args = options.pop('runner', '').split() + args
    extra_env = make_env(options)
    prepare_core_dump()
    print_command(args, extra_env)
    env = os.environ.copy()
    env.update(extra_env)
    return subprocess.Popen(args, env=env)


def use_memcheck(options, output='memcheck.out'):
    '''
    Set options to run through valgrind memcheck.
    WARNING valgrind does not handle vector instructions well,
    so try compiling without -march=native.
    '''
    suppressions = os.path.join(SRC, 'zmq.valgrind.suppressions')
    options = options.copy()
    options['runner'] = ' '.join([
        'valgrind',
        '--leak-check=full',
        '--show-reachable=yes',
        '--track-origins=yes',
        '--log-file={}'.format(output),
        '--suppressions={}'.format(suppressions),
    ])
    return options


def coverity():
    '''
    See http://scan.coverity.com
    '''
    buildtype = 'Debug'
    buildflag = '-DCMAKE_BUILD_TYPE={}'.format(buildtype)
    for d in [BUILD, COVERITY]:
        if os.path.exists(d):
            shutil.rmtree(d)
        os.makedirs(d)
    with chdir(BUILD):
        check_call('cmake', buildflag, ROOT)
        check_call('cov-build', '--dir', COVERITY, 'make')



from pomagma.compiler.expressions import Expression
from pomagma.compiler.util import inputs
from pomagma.compiler.parser import parse_corpus
from pomagma.util import TODO

HOLE = Expression.make('HOLE')


class Corpus(object):
    '''
    Corpus is a general-recursive set of definitions with holes.
    '''
    def __init__(self):
        self._defs = {}

    def load(self, filename):
        with open(filename) as f:
            self._defs = parse_corpus(f, filename=filename)

    def dump(self, filename):
        with open(filename, 'w') as f:
            f.write('# Corpus written by {}'.format(__file__))
            for var, expr in sorted(self._defs.iteritems()):
                assert isinstance(var, basestring), var
                assert isinstance(expr, Expression), expr
                f.write('\nEQUAL {} {}'.format(var, expr))

    def __getitem__(self, var):
        assert isinstance(var, Expression), var
        assert var.is_var(), var
        return self._defs.getitem(var, HOLE)

    def __setitem__(self, var, expr):
        assert isinstance(var, Expression), var
        assert isinstance(expr, Expression), expr
        assert var.is_var(), var
        if expr is HOLE:
            self._defs.pop(var, None)
        else:
            self._defs[var] = expr

    def __iter__(self):
        return self._defs.iteritems()

    def insert(self, key, value='HOLE'):
        var = Expression.make(key)
        expr = Expression.make(value)
        self[var] = expr


@inputs(Corpus)
def assess(corpus):
    TODO('report belief of validity of corpus')


@inputs(Corpus)
def refine(corpus, result_count=2**5, search_count=2**15):
    TODO('propose candidate joint hole-filling refinements of a corpus')



import re
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import Expression_1
from pomagma.compiler.frontend import write_full_programs
from pomagma.compiler.parser import parse_string_to_expr
from pomagma.compiler.parser import parse_theory_string
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.sugar import desugar_expr
from pomagma.compiler.sugar import desugar_sequent
from pomagma.compiler.util import memoize_arg
from pomagma.compiler.util import set_with

VAR = Expression_1('VAR')
RETURN = Expression_1('RETURN')
NRETURN = Expression_1('NRETURN')
NONEGATE = Expression_1('NONEGATE')


@memoize_arg
def guard_vars(expr):
    '''
    Whereas pomagma.compiler follows the variable naming convention defined in
    pomagma.compiler.signature.is_var(), pomagma.analyst.validate
    and the puddle editor require VAR guarded variables.
    '''
    if expr.name == 'VAR':
        return expr
    elif expr.is_var():
        return VAR(expr)
    else:
        args = map(guard_vars, expr.args)
        return Expression.make(expr.name, *args)


def desugar(string):
    expr = parse_string_to_expr(string)
    expr = desugar_expr(expr)
    expr = guard_vars(expr)
    return str(expr)


def compile_solver(expr, theory):
    '''
    Produces a set of programs that finds values of term satisfying a theory.
    Inputs:
        expr - string, an expression with free variables
        theory - string representing a theory (in .theory format)
    Outputs:
        a program to be consumed by the virtual machine
    Example:
        expr = 's'
        theory = """
            # 6 constraints = 4 facts + 2 rules
            LESS APP V s s       NLESS x BOT      NLESS x I
            LESS APP s BOT BOT   --------------   ----------------
            EQUAL APP s I I      LESS I APP s x   LESS TOP APP s x
            LESS TOP APP s TOP
            """
    '''
    assert isinstance(expr, basestring), expr
    assert isinstance(theory, basestring), theory
    expr = desugar_expr(parse_string_to_expr(expr))
    assert expr.vars, expr
    theory = parse_theory_string(theory)
    facts = map(desugar_expr, theory['facts'])
    rules = map(desugar_sequent, theory['rules'])
    sequents = []
    can_infer_necessary = not rules and all(f.vars <= expr.vars for f in facts)
    can_infer_possible = expr.is_var()  # TODO generalize to injective exprs
    if can_infer_necessary:
        sequents.append(Sequent(facts, [NONEGATE(RETURN(expr))]))
    if can_infer_possible:
        fail = NONEGATE(NRETURN(expr))
        sequents += [Sequent([], [f, fail]) for f in facts]
        sequents += [
            Sequent(r.antecedents, set_with(r.succedents, fail))
            for r in rules
        ]
    assert sequents, 'Cannot infer anything'
    programs = []
    write_full_programs(programs, sequents, can_parallelize=False)
    program = '\n'.join(programs)
    assert not re.search('FOR_BLOCK', program), 'cannot parallelize'
    return program



from pomagma.compiler import parser
from pomagma.compiler.util import find_theories


def test_parse_theory_file():
    for filename in find_theories():
        yield parser.parse_theory_file, filename


def _test_parse_theory_string(filename):
    with open(filename) as f:
        parser.parse_theory_string(f.read())


def test_parse_theory_string():
    for filename in find_theories():
        yield _test_parse_theory_string, filename



import os
from subprocess import CalledProcessError
import pomagma.util
from pomagma.analyst.client import Client

BINARY = os.path.join(pomagma.util.BIN, 'analyst', 'analyst')


class Server(object):

    def __init__(self, theory, world, address, **opts):
        language_file = os.path.join(
            pomagma.util.LANGUAGE,
            '{}.language'.format(theory))
        args = [
            BINARY,
            pomagma.util.abspath(world),
            pomagma.util.abspath(language_file),
            address.replace('tcp://localhost', 'tcp://*'),
        ]
        assert isinstance(address, basestring), address
        assert os.path.exists(world), world
        assert os.path.exists(language_file), language_file
        self._theory = theory
        self._address = address
        self._dir = os.path.abspath(os.curdir)
        self._log_file = pomagma.util.get_log_file(opts)
        self._proc = pomagma.util.log_Popen(*args, **opts)

    @property
    def theory(self):
        return self._theory

    @property
    def address(self):
        return self._address

    @property
    def pid(self):
        return self._proc.pid

    def connect(self):
        return Client(self.address, poll_callback=self.check)

    def stop(self):
        if self._proc.poll() is None:
            self._proc.terminate()

    def kill(self):
        self._proc.kill()

    def wait(self):
        if self._proc.wait() != 0:
            self.log_error()

    def check(self):
        if self._proc.poll() is not None:
            self.log_error()

    def log_error(self):
        pomagma.util.print_logged_error(self._log_file)
        with pomagma.util.chdir(self._dir):
            trace = pomagma.util.get_stack_trace(BINARY)
        pomagma.util.log_print(trace, self._log_file)
        raise CalledProcessError(self._proc.poll(), BINARY)



import os
from pomagma.compiler import sequents
from pomagma.compiler import parser
from pomagma.compiler.util import find_theories
from pomagma.compiler.sugar import desugar_theory

RULE_SETS = {
    os.path.basename(f): desugar_theory(parser.parse_theory_file(f))['rules']
    for f in find_theories()
}


def _test_contrapositives(name):
    print '# contrapositives'
    print
    for rule in RULE_SETS[name]:
        print rule.ascii()
        print
        if len(rule.succedents) != 1:
            print '    TODO'
            print
            continue
        for seq in sequents.get_contrapositives(rule):
            print seq.ascii(indent=4)
            print


def test_contrapositives():
    for name in RULE_SETS:
        yield _test_contrapositives, name


def _test_get_atomic(name):
    print '# get_atomic'
    print
    for rule in RULE_SETS[name]:
        print rule.ascii()
        print
        if len(rule.succedents) != 1:
            print '    TODO'
            print
            continue
        for seq in sequents.get_atomic(rule):
            print seq.ascii(indent=4)
            print


def test_get_atomic():
    for name in RULE_SETS:
        yield _test_get_atomic, name


def _test_normalize(name):
    print '# normalized'
    print
    for rule in RULE_SETS[name]:
        print rule.ascii()
        print
        if len(rule.succedents) != 1:
            print '    TODO'
            print
            continue
        for seq in sequents.normalize(rule):
            print seq.ascii(indent=4)
            print
            sequents.assert_normal(seq)


def test_normalize():
    for name in RULE_SETS:
        yield _test_normalize, name



from itertools import izip
from parsable import parsable
parsable = parsable.Parsable()
import pomagma.util


# ----------------------------------------------------------------------------
# Term enumeration


BOT, TOP = 'BOT', 'TOP'
I, K, F, B, C, W, S, Y = 'I', 'K', 'F', 'B', 'C', 'W', 'S', 'Y'
J, R, U, V, P = 'J', 'R', 'U', 'V', 'P'
UNKNOWNS = [
    'A', 'SECTION', 'RETRACT',
    'DIV', 'UNIT', 'SEMI', 'BOOL', 'BOOOL', 'NAT',
]


def iter_terms(atoms, max_atom_count):
    assert max_atom_count > 0
    for atom_count in xrange(1, 1 + max_atom_count):
        if atom_count == 1:
            for atom in atoms:
                yield (atom,)
        else:
            for lhs_count in xrange(1, atom_count):
                rhs_count = atom_count - lhs_count
                for lhs in iter_terms(atoms, lhs_count):
                    for rhs in iter_terms(atoms, rhs_count):
                        yield lhs + (rhs,)
    raise StopIteration()


# ----------------------------------------------------------------------------
# Convergence testing


class Converged(Exception):
    pass


class Diverged(Exception):
    pass


class Unknown(Exception):
    pass


def converge_step(term):
    head = term[0]
    assert isinstance(head, str), 'bad head: {}'.format(head)
    argv = term[1:]
    argc = len(argv)
    if head == TOP:
        raise Converged()
    elif head == BOT:
        raise Diverged()
    elif head == I:
        if argc == 0:
            return (TOP,)
        else:
            return argv[0] + argv[1:]
    elif head == K:
        if argc == 0:
            return (TOP,)
        else:
            return argv[0] + argv[2:]
    elif head == F:
        if argc == 0:
            return (TOP,)
        elif argc == 1:
            return (TOP,)
        else:
            return argv[1] + argv[2:]
    elif head == B:
        if argc == 0:
            return (TOP,)
        elif argc == 1:
            return argv[0]
        elif argc == 2:
            return argv[0] + (argv[1] + ((TOP,),),) + argv[3:]
        else:
            return argv[0] + (argv[1] + (argv[2],),) + argv[3:]
    elif head == C:
        if argc == 0:
            return (TOP,)
        elif argc == 1:
            return argv[0]
        elif argc == 2:
            return argv[0] + ((TOP,), argv[1],)
        else:
            return argv[0] + (argv[2], argv[1],) + argv[3:]
    elif head == W:
        if argc == 0:
            return (TOP,)
        elif argc == 1:
            return argv[0] + ((TOP,), (TOP,),)
        else:
            return argv[0] + (argv[1], argv[1],) + argv[2:]
    elif head == S:
        if argc == 0:
            return (TOP,)
        elif argc == 1:
            return argv[0] + ((TOP,), (TOP,),)
        elif argc == 2:
            return argv[0] + ((TOP,), argv[1] + ((TOP,),),)
        else:
            return argv[0] + (argv[2], argv[1] + (argv[2],),) + argv[3:]
    elif head == Y:
        if argc == 0:
            return (TOP,)
        else:
            return argv[0] + ((Y, argv[0],),) + argv[1:]
    elif head == J:
        if argc <= 1:
            return (TOP,)
        elif argc == 2:
            return (J, converge_step(argv[1]), argv[0],)
        else:
            return (J, argv[0] + argv[2:], argv[1] + argv[2:],)
    elif head == U:
        if argc == 0:
            return (TOP,)
        else:
            f = argv[0]
            return (J, f, (B, f, (U, f,),),) + argv[1:]
    elif head == V:
        if argc == 0:
            return (TOP,)
        else:
            f = argv[0]
            return (J, (I,), (B, f, (U, f,),),) + argv[1:]
    elif head == P:
        if argc <= 1:
            return (TOP,)
        else:
            f = argv[0]
            g = argv[1]
            return (V, (J, f, g,),) + argv[2:]
    elif head in [R]:
        raise Unknown()
    elif head in UNKNOWNS:
        raise Unknown()
    else:
        print 'WARNING unrecognized atom: {}'.format(head)
        raise Unknown()


def trivially_less(lhs, rhs):
    return lhs == (BOT,) or lhs == rhs or rhs == (TOP,)


def converge_less(lhs, rhs):
    if lhs == rhs:
        return True
    if len(rhs) < len(lhs):
        rhs = rhs + ((TOP,),) * (len(lhs) - len(rhs))
    if len(lhs) < len(rhs):
        lhs = lhs + ((TOP,),) * (len(rhs) - len(lhs))
    lhs_head, lhs_args = lhs[:1], lhs[1:]
    rhs_head, rhs_args = rhs[:1], rhs[1:]
    if not trivially_less(lhs_head, rhs_head):
        return False
    for lhs_arg, rhs_arg in izip(lhs_args, rhs_args):
        if not trivially_less(lhs_arg, rhs_arg):
            return False
    return True


def try_converge(term, steps):
    '''
    If a term 'DIV x' head reduces to a term less than or equal to itself,
    then it diverges.
    '''
    seen = set([term])
    for _ in xrange(steps):
        try:
            term = converge_step(term)
        except Unknown:
            return
        for other in seen:
            if converge_less(term, other):
                raise Diverged
        seen.add(term)


# ----------------------------------------------------------------------------
# Parsing


def parse_tokens_unsafe(tokens):
    head = tokens.pop()
    if head == 'APP':
        lhs = parse_tokens_unsafe(tokens)
        rhs = parse_tokens_unsafe(tokens)
        return lhs + (rhs,)
    elif head == 'COMP':
        lhs = parse_tokens_unsafe(tokens)
        rhs = parse_tokens_unsafe(tokens)
        return (B, lhs, rhs,)
    elif head == 'JOIN':
        lhs = parse_tokens_unsafe(tokens)
        rhs = parse_tokens_unsafe(tokens)
        return (J, lhs, rhs,)
    elif head == 'RAND':
        lhs = parse_tokens_unsafe(tokens)
        rhs = parse_tokens_unsafe(tokens)
        return (R, lhs, rhs,)
    elif head == 'CI':
        return (C, (I,),)
    elif head == 'CB':
        return (C, (B,),)
    else:
        return (head,)


def parse_tokens(tokens):
    term = parse_tokens_unsafe(tokens)
    assert not tokens, 'unexpected tokens: {}'.format(' '.join(tokens))
    return term


def parse_term(string):
    tokens = string.split()
    tokens.reverse()
    return parse_tokens(tokens)


def add_tokens(tokens, term):
    head, args = term[0], term[1:]
    tokens += ['APP'] * len(args)
    tokens.append(head)
    for arg in args:
        add_tokens(tokens, arg)


def print_term(term):
    tokens = []
    add_tokens(tokens, term)
    return ' '.join(tokens)


def stripped_lines(file_in):
    with open(file_in) as f:
        for line in f:
            line = line.split('#')[0].strip()
            if line:
                yield line
    raise StopIteration()


# ----------------------------------------------------------------------------
# Main


def try_prove_diverge(
        conjectures_in,
        conjectures_out,
        theorems_out,
        max_steps=20,
        log_file=None,
        log_level=0,
        **unused):
    assert conjectures_in != theorems_out
    assert conjectures_out != theorems_out

    def log_print(message):
        if log_file:
            pomagma.util.log_print(message, log_file)
        else:
            print message

    lines = list(stripped_lines(conjectures_in))
    log_print('Trying to prove {} conjectures'.format(len(lines)))

    conjecture_count = 0
    diverge_count = 0
    converge_count = 0
    with open(conjectures_out, 'w') as conjectures:
        conjectures.write('# divergence conjectures filtered by pomagma\n')
        with open(theorems_out, 'a') as theorems:

            def write_theorem(theorem):
                if log_level >= pomagma.util.LOG_LEVEL_DEBUG:
                    log_print('proved {}'.format(theorem))
                if diverge_count + converge_count == 0:
                    theorems.write('# divergence theorems proved by pomagma\n')
                theorems.write(theorem)
                theorems.write('\n')

            for line in lines:
                assert line.startswith('EQUAL BOT ')
                term_string = line[len('EQUAL BOT '):]
                term = parse_term(term_string)
                try:
                    try_converge(term, max_steps)
                    conjectures.write(line)
                    conjectures.write('\n')
                    conjecture_count += 1
                except Diverged:
                    theorem = 'EQUAL BOT {}'.format(term_string)
                    write_theorem(theorem)
                    diverge_count += 1
                except Converged:
                    theorem = 'NLESS {} BOT'.format(term_string)
                    write_theorem(theorem)
                    converge_count += 1
    if log_level >= pomagma.util.LOG_LEVEL_INFO:
        log_print('Proved {} diverge theorems'.format(diverge_count))
        log_print('Proved {} converge theorems'.format(converge_count))
        log_print('Failed to prove {} conjectures'.format(conjecture_count))
    theorem_count = diverge_count + converge_count
    return theorem_count


# ----------------------------------------------------------------------------
# Commands


@parsable
def print_terms(atoms='x,y', max_atom_count=3):
    '''
    Print all terms up to some max atom count.
    atoms is a comma-delimited list of atoms.
    '''
    atoms = atoms.split(',')
    for term in iter_terms(atoms, max_atom_count):
        print print_term(term)


@parsable
def count_terms(max_count=8):
    '''
    Count all terms up to some max atom count.
    '''
    atom_counts = range(1, 1 + max_count)
    max_counts = range(1, 1 + max_count)

    def count(a, m):
        return sum(1 for term in iter_terms(range(a), m))

    print '\t' * (max_count / 2) + '|atoms|'
    print '\t'.join(['|term|'] + [str(a) for a in atom_counts])
    print '-' * 8 * (1 + max_count)
    for m in max_counts:
        counts = [count(a, m) for a in atom_counts if a + m <= max_count + 1]
        print '\t'.join(map(str, [m] + counts))


@parsable
def may_diverge(atoms='I,K,F,B,C,W,S,Y', max_atom_count=4, max_steps=20):
    '''
    Print terms that have not been proven to converge.
    '''
    atoms = atoms.split(',')
    for term in iter_terms(atoms, max_atom_count):
        try:
            try_converge(term, max_steps)
            print term
        except Diverged:
            print term
        except Converged:
            pass


@parsable
def must_diverge(atoms='I,K,F,B,C,W,S,Y', max_atom_count=4, max_steps=20):
    '''
    Print terms that have been proven to diverge.
    '''
    atoms = atoms.split(',')
    for term in iter_terms(atoms, max_atom_count):
        try:
            try_converge(term, max_steps)
        except Converged:
            pass
        except Diverged:
            print term


if __name__ == '__main__':
    parsable()



from google.protobuf import text_format
from nose.tools import assert_equal
from pomagma.io import protobuf_test_pb2
from pomagma.io.protobuf import InFile
from pomagma.io.protobuf import OutFile
from pomagma.util import in_temp_dir
from pomagma.util.testing import for_each


def parse(text, Message=protobuf_test_pb2.TestMessage):
    message = Message()
    text_format.Merge(text, message)
    return message


EXAMPLES = [
    parse(''),
    parse('''
        optional_string: 'test'
    '''),
    parse('''
        repeated_string: 'test1'
        repeated_string: 'test2'
    '''),
    parse('''
        optional_string: 'test'
        repeated_string: 'test1'
        repeated_string: 'test2'
        optional_message: {
            repeated_message: {}
            repeated_message: {
                optional_string: 'sub sub 1'
                repeated_string: 'sub'
            }
            repeated_message: {
                optional_string: 'sub 1'
            }
            repeated_message: {
                repeated_string: 'sub 2'
            }
        }
    '''),
]


@for_each(EXAMPLES)
def test_write_read(expected):
    print 'Testing read(write({}))'.format(expected)
    actual = protobuf_test_pb2.TestMessage()
    with in_temp_dir():
        filename = 'test.pb'
        with OutFile(filename) as f:
            f.write(expected)
        with InFile(filename) as f:
            f.read(actual)
    assert_equal(actual, expected)



'''
Wrapper for Amazon S3 and compression.

References:
http://boto.readthedocs.org/en/latest/ref/s3.html
https://github.com/boto/boto/blob/develop/boto/s3
'''

import boto
import multiprocessing
import os
from parsable import parsable
import pomagma.util
import re
import subprocess

parsable = parsable.Parsable()


def try_connect_s3(bucket):
    if pomagma.util.TRAVIS_CI:
        print 'WARNING avoid connecting to bucket on travis-ci'
        return None
    try:
        connection = boto.connect_s3().get_bucket(bucket)
        print 'connected to bucket', bucket
        return connection
    except boto.exception.NoAuthHandlerFound as e:
        print 'WARNING failed to authenticate s3 bucket {}\n'.format(bucket), e
        return None
    except Exception as e:
        print 'WARNING failed to connect to s3 bucket {}\n'.format(bucket), e
        return None


# TODO allow different buckets to be specified, e.g. for a logging bucket
BUCKET = try_connect_s3(os.environ.get('POMAGMA_BUCKET', 'pomagma'))


def s3_lazy_put(filename, assume_immutable=False):
    '''
    Put file to s3 only if out of sync.
    '''
    key = BUCKET.get_key(filename)
    if key is None:
        key = BUCKET.new_key(filename)
        print 'uploading', filename
        key.set_contents_from_filename(filename)
        return key
    elif assume_immutable:
        # print 'already synchronized'
        return key
    else:
        with open(filename, 'rb') as f:
            # print 'checking cached', filename
            md5 = key.compute_md5(f)
        key_md5_0 = key.etag.strip('"')  # WTF
        if md5[0] == key_md5_0:
            # print 'already synchronized'
            return key
        else:
            print 'uploading', filename
            key.set_contents_from_filename(filename, md5=md5)
            return key


def s3_lazy_get(filename, assume_immutable=False):
    '''
    Get file from s3 only if out of sync.
    '''
    key = BUCKET.get_key(filename)
    if key is None:
        print 'missing', filename
        return key
    if os.path.exists(filename):
        if assume_immutable:
            # print 'already synchronized'
            return key
        else:
            with open(filename, 'rb') as f:
                # print 'checking cached', filename
                md5 = key.compute_md5(f)
            key_md5_0 = key.etag.strip('"')  # WTF
            if md5[0] == key_md5_0:
                # print 'already synchronized'
                return key
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    print 'downloading', filename
    key.get_contents_to_filename(filename)
    return key


def s3_listdir(prefix):
    keys = BUCKET.list(prefix)
    return [key for key in keys if not key.name.endswith('/')]


def s3_remove(filename):
    key = BUCKET.get_key(filename)
    if key is not None:
        key.delete()


def s3_exists(filename):
    key = BUCKET.get_key(filename)
    return key is not None


def bzip2(filename):
    subprocess.check_call(['bzip2', '--keep', '--force', filename])
    filename_ext = filename + '.bz2'
    return filename_ext


def bunzip2(filename_ext):
    assert filename_ext[-4:] == '.bz2', filename_ext
    filename = filename_ext[:-4]
    subprocess.check_call(['bunzip2', '--keep', '--force', filename_ext])
    return filename


def _silent_check_call(args):
    with open(os.devnull, 'w') as devnull:
        subprocess.check_call(args, stderr=devnull, stdout=devnull)


def archive_7z(filename):
    filename_ext = filename + '.7z'
    if os.path.exists(filename_ext):
        os.remove(filename_ext)
    print 'compressing', filename
    _silent_check_call(['7z', 'a', '-y', filename_ext, filename])
    return filename_ext


def extract_7z(filename_ext):
    assert filename_ext[-3:] == '.7z', filename_ext
    filename = filename_ext[:-3]
    if os.path.exists(filename):
        os.remove(filename)
    print 'extracting', filename_ext
    _silent_check_call(['7z', 'x', '-mtm=off', '-y', filename_ext])
    return filename


ARCHIVE = archive_7z
EXTRACT = extract_7z
EXT = '.7z'


# blobs are immutable and compressed
def is_blob(filename):
    if re.match('[a-z0-9]{40}$', os.path.basename(filename)):
        # This only works on local files:
        # mode = oct(os.stat(filename).st_mode)[-3:]
        # assert mode == '444', 'invalid blob mode: {}'.format(mode)
        return True
    else:
        return False


def get(filename):
    if is_blob(filename):
        s3_lazy_get(filename, assume_immutable=True)
        os.chmod(filename, 0444)
    else:
        filename_ext = filename + EXT
        s3_lazy_get(filename_ext)
        EXTRACT(filename_ext)


def put(filename):
    if is_blob(filename):
        s3_lazy_put(filename, assume_immutable=True)
    else:
        filename_ext = ARCHIVE(filename)
        s3_lazy_put(filename_ext)


def listdir(prefix=''):
    for key in s3_listdir(prefix):
        if is_blob(key.name):
            filename = key.name
        else:
            filename_ext = key.name
            assert filename_ext[-len(EXT):] == EXT, filename_ext
            filename = filename_ext[:-len(EXT)]
        yield filename


def remove(filename):
    if os.path.exists(filename):
        os.remove(filename)
    if is_blob(filename):
        s3_remove(filename)
    else:
        filename_ext = filename + EXT
        if os.path.exists(filename_ext):
            os.remove(filename_ext)
        s3_remove(filename_ext)


def parallel_map(fun, args):
    if len(args) <= 1:
        return map(fun, args)
    else:
        return multiprocessing.Pool().map(fun, args)


def filter_cache(filenames):
    return [
        f
        for f in filenames
        if f[-len(EXT):] != EXT
        if not os.path.exists(f) or os.path.isfile(f)
    ]


BLACKLIST = re.compile('(test|core|temp|mutex|queue|socket|7z)')


def find(path):
    if os.path.isdir(path):
        return filter(os.path.isfile, [
            os.path.abspath(os.path.join(root, filename))
            for root, dirnames, filenames in os.walk(path)
            if not BLACKLIST.search(root)
            for filename in filenames
            if not BLACKLIST.search(filename)
        ])
    elif not BLACKLIST.search(os.path.basename(path)):
        return [os.path.abspath(path)]
    else:
        return []


@parsable
def find_s3(prefix=''):
    '''
    Find copyable files on S3.
    '''
    assert BUCKET
    for filename in listdir(prefix):
        print filename


@parsable
def find_local(path='.'):
    '''
    Find copyable files on local filesystem.
    '''
    for filename in find(path):
        print filename


@parsable
def snapshot(source, destin):
    '''
    Create resursive snapshot of hard links for push/pull.
    '''
    assert os.path.isdir(source)
    assert destin != source
    source = os.path.abspath(source)
    destin = os.path.abspath(destin)
    for source_file in find(source):
        relpath = os.path.relpath(source_file, source)
        destin_file = os.path.join(destin, relpath)
        destin_dir = os.path.dirname(destin_file)
        if not os.path.exists(destin_dir):
            os.makedirs(destin_dir)
        if os.path.exists(destin_file):
            os.remove(destin_file)
        os.link(source_file, destin_file)


@parsable
def pull(*filenames):
    '''
    Pull files from S3 into local cache.
    '''
    assert BUCKET
    filenames = sum([
        list(listdir(f)) if f.endswith('/') else [f]
        for f in filenames
    ], [])
    parallel_map(get, filter_cache(filenames))


@parsable
def push(*filenames):
    '''
    Push files to S3 from local cache.
    '''
    assert BUCKET
    filenames = sum(map(find, filenames), [])
    filenames = map(os.path.relpath, filenames)
    parallel_map(put, filter_cache(filenames))


@parsable
def rm(*filenames):
    '''
    Remove files from S3 and local cache.
    '''
    assert BUCKET
    parallel_map(remove, filter_cache(filenames))


if __name__ == '__main__':
    parsable()



import math
from pomagma.compiler.expressions import Expression_1
from pomagma.compiler.util import log_sum_exp
from pomagma.compiler.util import memoize_make
from pomagma.compiler.util import set_with


def assert_in(element, set_):
    assert element in set_, (element, set_)


def assert_not_in(element, set_):
    assert element not in set_, (element, set_)


def assert_subset(subset, set_):
    assert subset <= set_, (subset, set_)


OBJECT_COUNT = 1e4                          # optimize for this many obs
LOGIC_COST = OBJECT_COUNT / 64.0            # perform logic on 64-bit words
LOG_OBJECT_COUNT = math.log(OBJECT_COUNT)

UNKNOWN = Expression_1('UNKNOWN')


def add_costs(costs):
    return (log_sum_exp(*(LOG_OBJECT_COUNT * c for c in costs)) /
            LOG_OBJECT_COUNT)


class Plan(object):
    __slots__ = ['_args', '_cost', '_rank']

    def __init__(self, *args):
        self._args = args
        self._cost = None
        self._rank = None

    @property
    def cost(self):
        if self._cost is None:
            self._cost = math.log(self.op_count()) / LOG_OBJECT_COUNT
        return self._cost

    @property
    def rank(self):
        if self._rank is None:
            s = repr(self)
            self._rank = self.cost, len(s), s
        return self._rank

    def __lt__(self, other):
        return self.rank < other.rank

    def permute_symbols(self, perm):
        return self.__class__.make(*(
            a.permute_symbols(perm)
            for a in self._args
        ))


@memoize_make
class Iter(Plan):
    __slots__ = ['_repr', 'var', 'body', 'tests', 'lets', 'stack']

    def __init__(self, var, body):
        Plan.__init__(self, var, body)
        assert var.is_var(), var
        assert isinstance(body, Plan), body
        self._repr = None
        self.var = var
        self.body = body
        self.tests = []
        self.lets = {}
        self.stack = set()
        self.optimize()

    def add_test(self, test):
        assert isinstance(test, Test), 'add_test arg is not a Test'
        self.tests.append(test.expr)
        self.stack.add(test)

    def add_let(self, let):
        assert isinstance(let, Let), 'add_let arg is not a Let'
        assert let.var not in self.lets, 'add_let var is not in Iter.lets'
        self.lets[let.var] = let.expr
        self.stack.add(let)

    def __repr__(self):
        if self._repr is None:
            # Optimized:
            # tests = ['if {}'.format(t) for t in self.tests]
            # lets = ['let {}'.format(l) for l in sorted(self.lets.iterkeys())]
            # self._repr = 'for {0}: {1}'.format(
            #     ' '.join([str(self.var)] + tests + lets),
            #     self.body)
            self._repr = 'for {}: {}'.format(self.var, self.body)
        return self._repr

    def validate(self, bound):
        assert_not_in(self.var, bound)
        bound = set_with(bound, self.var)
        for test in self.tests:
            assert_subset(test.vars, bound)
        for var, expr in self.lets.iteritems():
            assert_subset(expr.vars, bound)
            assert_not_in(var, bound)
        self.body.validate(bound)

    def op_count(self, stack=None):
        logic_cost = LOGIC_COST * (len(self.tests) + len(self.lets))
        object_count = OBJECT_COUNT
        for test_or_let in self.stack:
            object_count *= test_or_let.prob()
        let_cost = len(self.lets)
        body_cost = self.body.op_count(stack=self.stack)
        return logic_cost + object_count * (let_cost + body_cost)

    def optimize(self):
        node = self.body
        new_lets = set()
        while isinstance(node, Test) or isinstance(node, Let):
            if isinstance(node, Let):
                new_lets.add(node.var)
            expr = node.expr
            while expr.name == 'UNKNOWN':
                expr = expr.args[0]
            optimizable = (
                self.var in expr.vars and
                expr.vars.isdisjoint(new_lets) and
                sum(1 for arg in expr.args if self.var == arg) == 1 and
                sum(1 for arg in expr.args if self.var in arg.vars) == 1 and
                (isinstance(node, Let) or expr.is_rel())
            )
            if optimizable:
                if isinstance(node, Test):
                    self.add_test(node)
                else:
                    self.add_let(node)
            node = node.body


# TODO injective function inverse need not be iterated
@memoize_make
class IterInvInjective(Plan):
    __slots__ = ['fun', 'value', 'var', 'body']

    def __init__(self, fun, body):
        Plan.__init__(self, fun, body)
        assert fun.arity == 'InjectiveFunction'
        self.fun = fun.name
        self.value = fun.var
        (self.var,) = fun.args
        self.body = body

    def __repr__(self):
        return 'for {0} {1}: {2}'.format(self.fun, self.var, self.body)

    def validate(self, bound):
        assert_in(self.value, bound)
        assert_not_in(self.var, bound)
        self.body.validate(set_with(bound, self.var))

    def op_count(self, stack=None):
        return 4.0 + 0.5 * self.body.op_count()  # amortized


@memoize_make
class IterInvBinary(Plan):
    __slots__ = ['fun', 'value', 'var1', 'var2', 'body']

    def __init__(self, fun, body):
        Plan.__init__(self, fun, body)
        assert fun.arity in ['BinaryFunction', 'SymmetricFunction']
        self.fun = fun.name
        self.value = fun.var
        self.var1, self.var2 = fun.args
        self.body = body

    def __repr__(self):
        return 'for {0} {1} {2}: {3}'.format(
            self.fun, self.var1, self.var2, self.body)

    def validate(self, bound):
        assert_in(self.value, bound)
        assert_not_in(self.var1, bound)
        assert_not_in(self.var2, bound)
        self.body.validate(set_with(bound, self.var1, self.var2))

    def op_count(self, stack=None):
        return 4.0 + 0.25 * OBJECT_COUNT * self.body.op_count()  # amortized


@memoize_make
class IterInvBinaryRange(Plan):
    __slots__ = ['fun', 'value', 'var1', 'var2', 'lhs_fixed', 'body']

    def __init__(self, fun, fixed, body):
        Plan.__init__(self, fun, fixed, body)
        assert fun.arity in ['BinaryFunction', 'SymmetricFunction']
        self.fun = fun.name
        self.value = fun.var
        self.var1, self.var2 = fun.args
        assert self.var1 != self.var2
        assert self.var1 == fixed or self.var2 == fixed
        self.lhs_fixed = (fixed == self.var1)
        self.body = body

    def __repr__(self):
        if self.lhs_fixed:
            return 'for {0} ({1}) {2}: {3}'.format(
                self.fun, self.var1, self.var2, self.body)
        else:
            return 'for {0} {1} ({2}): {3}'.format(
                self.fun, self.var1, self.var2, self.body)

    def validate(self, bound):
        assert self.value in bound
        if self.lhs_fixed:
            assert_in(self.var1, bound)
            assert_not_in(self.var2, bound)
            self.body.validate(set_with(bound, self.var2))
        else:
            assert_in(self.var2, bound)
            assert_not_in(self.var1, bound)
            self.body.validate(set_with(bound, self.var1))

    def op_count(self, stack=None):
        return 4.0 + 0.5 * self.body.op_count()  # amortized


@memoize_make
class Let(Plan):
    __slots__ = ['var', 'expr', 'body']

    def __init__(self, expr, body):
        Plan.__init__(self, expr, body)
        assert isinstance(body, Plan)
        assert expr.is_fun()
        self.var = expr.var
        self.expr = expr
        self.body = body

    def __repr__(self):
        return 'let {0}: {1}'.format(self.var, self.body)

    def validate(self, bound):
        assert_subset(self.expr.vars, bound)
        assert_not_in(self.var, bound)
        self.body.validate(set_with(bound, self.var))

    __probs = {'NullaryFunction': 0.9}

    def prob(self):
        return self.__probs.get(self.expr.arity, 0.1)

    def op_count(self, stack=None):
        if stack and self in stack:
            return self.body.op_count(stack=stack)
        else:
            return 1.0 + self.prob() * self.body.op_count(stack=stack)


@memoize_make
class Test(Plan):
    __slots__ = ['expr', 'body']

    def __init__(self, expr, body):
        Plan.__init__(self, expr, body)
        assert not expr.is_var()
        assert isinstance(body, Plan)
        self.expr = expr
        self.body = body

    def __repr__(self):
        return 'if {0}: {1}'.format(self.expr, self.body)

    def validate(self, bound):
        assert_subset(self.expr.vars, bound)
        self.body.validate(bound)

    __probs = {'NLESS': 0.9}

    def prob(self):
        return self.__probs.get(self.expr.name, 0.1)

    def op_count(self, stack=None):
        if stack and self in stack:
            return self.body.op_count(stack=stack)
        else:
            return 1.0 + self.prob() * self.body.op_count(stack=stack)


@memoize_make
class Ensure(Plan):
    __slots__ = ['expr']

    def __init__(self, expr):
        Plan.__init__(self, expr)
        assert expr.args, ('expr is not compound', expr)
        self.expr = expr

    def __repr__(self):
        return 'ensure {0}'.format(self.expr)

    def validate(self, bound):
        assert_subset(self.expr.vars, bound)

    def op_count(self, stack=None):
        fun_count = 0
        if self.expr.name == 'EQUATION':
            for arg in self.expr.args:
                if arg.is_fun():
                    fun_count += 1
        return [1.0, 1.0 + 0.5 * 1.0, 2.0 + 0.75 * 1.0][fun_count]



import os
import glob
from nose.tools import assert_almost_equal
from nose.tools import assert_set_equal
from .util import dict_to_language
from .util import json_load
from .util import language_to_dict
from .util import normalize_dict


SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def assert_converts(expected):
    language = dict_to_language(expected)
    actual = language_to_dict(language)
    assert_set_equal(set(expected.keys()), set(actual.keys()))
    for arity in expected.keys():
        assert_set_equal(
            set(expected[arity].keys()),
            set(actual[arity].keys()))
        for term, weight in expected[arity].iteritems():
            assert_almost_equal(weight, actual[arity][term])


def test_example():
    example = {
        'NULLARY': {
            'B': 1.0,
            'C': 1.30428,
            'I': 2.21841,
            'K': 2.6654,
            'S': 2.69459,
        },
        'BINARY': {
            'APP': 0.4,
            'COMP': 0.2,
        },
    }
    normalize_dict(example)
    assert_converts(example)


def test_convert_json():
    for filename in glob.glob(os.path.join(SCRIPT_DIR, '*.json')):
        example = json_load(filename)
        normalize_dict(example)
        yield assert_converts, example



import pomagma.util
from pomagma.compiler import __main__ as main
from pomagma.compiler.util import find_theories


def _test_compile(filename):
    with pomagma.util.in_temp_dir():
        main.compile(
            filename,
            symbols_out='temp.symbols',
            facts_out='temp.facts',
            programs_out='temp.programs',
            optimized_out='temp.optimized.programs')


def test_compile_rules():
    for filename in find_theories():
        yield _test_compile, filename


def test_compile_facts():
    for filename in find_theories():
        yield _test_compile, filename



import contextlib
import pomagma.util
from pomagma.cartographer import server

serve = server.Server


@contextlib.contextmanager
def load(theory, world, address=None, **opts):
    with pomagma.util.log_duration():
        server = serve(theory, world, address, **opts)
        with server.connect() as client:
            yield client
            client.stop()
        server.wait()



from pomagma.compiler.parser import parse_string_to_expr
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.util import cached
from pomagma.compiler.util import methodof
import pomagma.compiler.syntax as proto


@methodof(proto.Expression)
def pythonize_expr(proto_expr):
    return parse_string_to_expr(proto_expr.polish)


@methodof(proto.Sequent)
def pythonize_sequent(proto_sequent):
    antecedents = map(pythonize_expr, proto_sequent.antecedents)
    succedents = map(pythonize_expr, proto_sequent.succedents)
    return Sequent(antecedents, succedents)


@methodof(proto.Theory)
def pythonize_theory(proto_theory):
    return {
        'facts': set(map(pythonize_expr, proto_theory.facts)),
        'rules': set(map(pythonize_sequent, proto_theory.rules)),
    }


@cached
@methodof(proto.Theory, proto.Theory)
def entails(lhs, rhs):
    raise NotImplementedError()



from pomagma.io import creat
import gzip


class InFile(object):
    def __init__(self, filename):
        self._gzip = gzip.GzipFile(filename, 'rb')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self._gzip.close()

    def read(self, message):
        message.ParseFromString(self._gzip.read())


class OutFile(object):
    def __init__(self, filename):
        self._file = creat(filename, 0444)
        self._gzip = gzip.GzipFile(mode='wb', fileobj=self._file)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self._gzip.close()
        self._file.close()

    def write(self, message):
        self._gzip.write(message.SerializeToString())
